# -*- coding: utf-8 -*-
from typing import List, Iterable
import sympy as sy
from sympy import latex
import numpy as np


def gen_Lagrange_1d(*args, x:Iterable=None, i:List[int]=None, xsym: str=None, 
                    fsym: str=None, sym:bool=False, N:int=None, **kwargs) -> dict:
    r"""
    Generates Lagrange polynomials and their derivatives up to 3rd, for appeoximation 
    in 1d space, based on N input pairs of position and value. Geometrical parameters 
    can be numeric or symbolic.

    Parameters
    ----------        
    x : Iterable, Optional
        The locations of the data points. If not specified and `sym=False`, a range 
        of [-1, 1] is assumed and the locations are generated as `np.linspace(-1, 1, N)`, 
        where N is the number of data points. If `sym=True`, the calculation is entirely
        symbolic. Default is None.

    i : List[int]
        If not specified, indices are assumed as [1, ..., N], but this is only relevant 
        for symbolic representation, not the calculation itself, which only cares about 
        the number of data points, regardless of their actual indices.
    
    xsym : str, Optional
        Symbol of the variable in the symbolic representation of the generated functions.
        Default is :math:`x`.

    fsym : str, Optional
        Symbol of the function in the symbolic representation of the generated functions.
        Default is 'f'.

    sym : bool, Optional.
        If True, locations of the data points are left in a symbolic state. This requires
        the inversion of a symbolic matrix, which has some reasonable limitations.
        Default is False.
        
    N : int, Optional
        If neither 'x' nor 'i' is specified, this controls the number of functions to generate.
        Default is None.

    Returns
    -------
    dict
        A dictionary containing the generated functions for the reuested nodes.
        The keys of the dictionary are the indices of the points, the values are
        dictionaries with the following keys and values:

            symbol : the sympy symbol of the function

            0 : the function

            1 : the first derivative as a sympy expression

            2 : the second derivative as a sympy expression

            3 : the third derivative as a sympy expression
            
    Example
    -------
    >>> from neumann.approx.lagrange import gen_Lagrange_1d
    
    To generate approximation functions for a 2-noded line:
    
    >>> gen_Lagrange_1d(x=[-1, 1])
    
    or equivalently
    
    >>> gen_Lagrange_1d(N=2)
    
    To generate the same functions in symbolic form:
    
    >>> gen_Lagrange_1d(i=[1, 2], sym=True)

    Notes
    -----
    Inversion of a heavily symbolic matrix may take quite some time, and is not suggested
    for N > 3. This is one reason why isoparametric finite elements make sense. 
    Fixing the locations as constant real numbers symplifies the process and makes 
    the solution much faster.

    """
    xsym = xsym if xsym is not None else r'x'
    fsym = fsym if fsym is not None else r'\phi'
    module_data = {}
    xvar = sy.symbols(xsym)
    if not isinstance(N, int):
        N = len(x) if x is not None else len(i)
    inds = list(range(1, N + 1)) if i is None else i
    def var_tmpl(i): return r'\Delta_{}'.format(i)
    def var_str(i): return var_tmpl(inds[i])
    coeffs = sy.symbols(', '.join(['c_{}'.format(i+1) for i in range(N)]))
    variables = sy.symbols(', '.join([var_str(i) for i in range(N)]))
    if x is None:
        if xsym is None or not sym:
            x = np.linspace(-1, 1, N)
        else:
            x = sy.symbols(
                ', '.join([xsym + '_{}'.format(i+1) for i in range(N)]))
    poly = sum([c * xvar**i for i, c in enumerate(coeffs)])
    #
    evals = [poly.subs({'x': x[i]}) for i in range(N)]
    A = sy.zeros(N, N)
    for i in range(N):
        A[i, :] = sy.Matrix([evals[i].coeff(c) for c in coeffs]).T
    coeffs_new = A.inv() * sy.Matrix(variables)
    subs = {coeffs[i]: coeffs_new[i] for i in range(N)}
    approx = poly.subs(subs).simplify().expand()
    #
    shp = [approx.coeff(v).factor().simplify() for v in variables]
    #
    def diff(fnc): return fnc.diff(
        xvar).expand().simplify().factor().simplify()
    dshp1 = [diff(fnc) for fnc in shp]
    dshp2 = [diff(fnc) for fnc in dshp1]
    dshp3 = [diff(fnc) for fnc in dshp2]
    #
    for i, ind in enumerate(inds):
        module_data[ind] = {}
        fnc_str = latex(sy.symbols(fsym + '_{}'.format(ind)))
        module_data[ind]['symbol'] = fnc_str
        module_data[ind][0] = shp[i]
        module_data[ind][1] = dshp1[i]
        module_data[ind][2] = dshp2[i]
        module_data[ind][3] = dshp3[i]
    return module_data


# -*- coding: utf-8 -*-


__all__ = ['DegenerateProblemError', 'NoSolutionError', 'PreProcError', 'PostProcError']


class SolutionError(Exception):
    """
    Raises during the solution stage of a solution.
    
    """
    pass


class DegenerateProblemError(SolutionError):
    """
    The problem is degenerate. 
    
    The objective could be decreased, but only on the expense
    of violating positivity of the standard variables.
    
    """
    pass


class NoSolutionError(SolutionError):
    """
    There is no solution to this problem.
    
    Step size could be indefinitely increased in a
    direction without violating feasibility.
    
    """
    pass


class PreProcError(Exception):
    """
    Raises during the postprocessing stage of a solution.
    
    """
    pass


class PostProcError(Exception):
    """
    Raises during the postprocessing stage of a solution.
    
    """
    pass


# -*- coding: utf-8 -*-
import numpy as np
from numpy import concatenate as join, ndarray

from awkward import unflatten, Array

def shape2d(arr): return arr.shape[:2]
def cut(shp): return np.full(shp[0], shp[1], dtype=np.int64)
def flatten(arr): return arr.flatten()


class JaggedArray(Array):
    """
    A numba-jittable class that handles matrices with a variable 
    number of columns per row.

    The class is actually an interface to `awkward.Array`,
    with some additional features, specific to 2d jagged arrays.

    Parameters
    ----------
    data : Iterable
        A 2d numpy array or a list of such objects.
    behaviour
        Passed forward to `awkward.Array`, see it's documentation
        for the details.
    with_name
        Passed forward to `awkward.Array`, see it's documentation
        for the details.
    check_valid
        Passed forward to `awkward.Array`, see it's documentation
        for the details.
    cache
        Passed forward to `awkward.Array`, see it's documentation
        for the details.
    kernels
        Passed forward to `awkward.Array`, see it's documentation
        for the details.
    cuts : Iterable
        An iterable that tells how to unflatten an 1d array into a 
        2d jagged shape.

    Returns
    -------
    JaggedArray

    """

    def __init__(self, data, behavior=None, with_name=None,
                 check_valid=False, cache=None, kernels=None,
                 cuts=None) -> 'JaggedArray':
        if isinstance(data, np.ndarray):
            nD = len(data.shape)
            assert nD <= 2, "Only 2 dimensional arrays are supported!"
            if nD == 1:
                assert isinstance(cuts, np.ndarray)
                data = unflatten(data, cuts)
            else:
                assert nD == 2
        elif isinstance(data, list):
            if isinstance(data[0], ndarray):
                # data is a list of 2d numpy arrays
                assert all(map(lambda arr: len(arr.shape) == 2, data)), \
                    "Only 2 dimensional arrays are supported!"
                # NOTE - implementaion 1
                # > Through the back door, but this is probably the cleanest solution of all.
                # > It only requires to create one python list, without further operations on it.
                data = unflatten(join(list(map(flatten, data))),
                                 join(list(map(cut, map(shape2d, data)))))
                # NOTE - implementaion 2
                #from operator import add
                #from functools import reduce
                # > This also works, but requires data to jump back and forth just to
                # > have a merged list of lists. It also requires to add nested python lists,
                # > which is probably not the quickest operation in the computational world.
                #data = ak.from_iter(reduce(add, map(lambda arr : ak.Array(arr).to_list(), data)))
                # NOTE - implementaion 3
                # > This also works at creation, but fails later at some operation due to
                # > the specific layout generated by ak.concatenate
                #data = ak.concatenate(list(map(lambda arr : ak.Array(arr), data)))
        super().__init__(data, behavior=behavior, with_name=with_name,
                         check_valid=check_valid, cache=cache, kernels=kernels)



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange
__cache = True


def lower_spdata(data: ndarray, rows: ndarray, cols: ndarray):
    inds = np.where(rows >= cols)[0]
    return data[inds], rows[inds], cols[inds]


def upper_spdata(data: ndarray, rows: ndarray, cols: ndarray):
    inds = np.where(cols >= rows)[0]
    return data[inds], rows[inds], cols[inds]


@njit(nogil=True, parallel=True, cache=__cache)
def get_shape_sp(indptr: np.ndarray):
    nE = len(indptr)-1
    widths = np.zeros(nE, dtype=indptr.dtype)
    for iE in prange(nE):
        widths[iE] = indptr[iE+1] - indptr[iE]
    return nE, widths.max()


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def count_cols(arr: np.ndarray):
    n = len(arr)
    res = np.zeros(n, dtype=np.int64)
    for i in prange(n):
        res[i] = len(arr[i])
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numba import jit
from numba.types import int64, Array
from numba.typed import Dict

from ..linalg.sparse import csr_matrix

__all__ = ['Graph', 'rooted_level_structure', 'pseudo_peripheral_nodes']

try:
    import networkx as ntx
    
    try:
        adjacency_matrix = ntx.to_scipy_sparse_array
    except Exception:
        adjacency_matrix = ntx.adjacency_matrix

    class Graph(ntx.Graph):
        """
        A subclass of `networkx.Graph`, extending its capabilities.
        See the documentation of `networkx` for the details on how to
        define graphs.

        Note
        ----
        If `networkx` is not installed, the class is `NoneType`, but the
        functionality it implements is still available, you just have to
        manage graph creation by yourself.

        Examples
        --------
        A basic example with `networkx`:

        >>> from neumann.topology import Graph
        >>> import networkx as nx
        >>> grid = nx.grid_2d_graph(5, 5)  # 5x5 grid
        >>> G = Graph(grid)

        """

        def adjacency_matrix(self, *args, to_csr:bool=False, **kwargs):
            """
            Returns the adjacency matrix of the graph.

            Parameters
            ----------
            to_csr : bool, Optional
                If `True`, the result of networkx.adjacency_matrix is 
                returned as a csr_matrix.

            *args : Tuple, Optional
                Forwarded to networkx.adjacency_matrix

            **kwargs, dict, Optional
                Forwarded to networkx.adjacency_matrix

            Returns
            -------
            NumPy array, SciPy array or csr_matrix
                The adjacency representation of the graph.

            Examples
            --------
            >>> from neumann.topology import Graph
            >>> G = Graph([(1, 1)])
            >>> A = G.adjacency_matrix()
            >>> print(A.todense())
            [[1]]

            """
            adj = adjacency_matrix(self, *args, **kwargs)
            return csr_matrix(adj) if to_csr else adj

        def rooted_level_structure(self, root=0):
            """
            Returns the rooted level structure (RLS) of the graph.

            The call is forwarded to `rooted_level_structure`, go there
            to read about the possible arguments.

            See Also
            --------
            :func:`rooted_level_structure`

            """
            return rooted_level_structure(csr_matrix(adjacency_matrix(self)), root)

        def pseudo_peripheral_nodes(self):
            """
            Returns the indices of nodes that are possible candidates
            for being peripheral nodes of a graph.
            
            """
            return pseudo_peripheral_nodes(csr_matrix(adjacency_matrix(self)))
except:
    Graph = None

int64A = Array(int64, 1, 'C')


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def rooted_level_structure(adj: csr_matrix, root: int = 0) -> Dict:
    """
    Turns a sparse adjacency matrix into a rooted level structure.

    Parameters
    ----------
    adj : csr_matrix
        Adjacency matrix in CSR format.

    root : int, Optional
        Index of the root node. Default is 0.

    Returns
    -------
    dict
        A `numba` dictionary <int[:] : int[:, :]>, where the keys
        refer to different levels, and the values are the indices
        of nodes on that level.

    """
    nN = len(adj.indptr) - 1
    rls = Dict.empty(
        key_type=int64,
        value_type=int64A,
    )
    level = 0
    rls[level] = np.array([root], dtype=np.int64)
    nodes = np.zeros(nN, dtype=np.int64)
    nodes[root] = 1
    levelset = np.zeros(nN, dtype=np.int64)
    nE = 1
    while nE < nN:
        levelset[:] = 0
        for node in rls[level]:
            _, neighbours = adj.row(node)
            levelset[neighbours] = 1
        for iN in range(nN):
            if nodes[iN] == 1:
                levelset[iN] = 0
        level += 1
        rls[level] = np.where(levelset == 1)[0]
        nE += len(rls[level])
        for iN in range(nN):
            if levelset[iN] == 1:
                nodes[iN] = 1
    return rls


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def pseudo_peripheral_nodes(adj: csr_matrix) -> np.ndarray:
    """
    Returns the indices of nodes that are possible candidates
    for being peripheral nodes of a graph.

    Parameters
    ----------
    adj : csr_matrix
        Adjacency matrix in CSR format.

    Returns
    -------
    numpy.ndarray
        Integer array of nodal indices.

    """

    def length_width(RLS):
        length = len(RLS)
        width = 0
        for i in range(length):
            width = max(width, len(RLS[i]))
        return length, width

    RLS = rooted_level_structure(adj, root=0)
    length, width = length_width(RLS)
    while True:
        nodes = RLS[len(RLS) - 1]
        found = False
        for _, node in enumerate(nodes):
            iRLS = rooted_level_structure(adj, root=node)
            iL, iW = length_width(iRLS)
            if (iL > length) or (iL == length and iW < width):
                RLS = iRLS
                length = iL
                width = iW
                found = True
        if not found:
            nR = len(RLS[len(RLS) - 1]) + 1
            res = np.zeros(nR, dtype=np.int64)
            res[:-1] = RLS[len(RLS) - 1]
            res[-1] = RLS[0][0]
            return res


# -*- coding: utf-8 -*-
import numpy as np

from .alphabet import latinrange

from .frame import ReferenceFrame as Frame
from .vector import Vector, VectorBase


class Tensor(Vector):

    def __init__(self, *args, symmetric=False, isotropic=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.symmetric = symmetric  # has no effect, just a reminder
        self.isotropic = isotropic  # has no effect, just a reminder

    @classmethod
    def identity(cls, dim=1):
        return cls(np.eye(dim))

    @staticmethod
    def eye(dim=1):
        return Tensor(np.eye(dim))

    def _transform(self, dcm: np.ndarray = None):
        Q = dcm.T
        dim = self.dim
        source = latinrange(dim, start='i')
        target = latinrange(dim, start=ord('i') + dim)
        command = ','.join([t + s for t, s in zip(target, source)]) + \
            ',' + ''.join(source)
        args = [Q for _ in range(dim)]
        return np.einsum(command, *args, self._array, optimize='greedy')




""  # -*- coding: utf-8 -*-
from numba import njit
import numpy as np
from scipy.special import factorial as fact
from numpy import outer
from collections.abc import Iterable
from numba import njit


from .func import isMLSWeightFunction, ConstantWeightFunction


def moving_least_squares(points, *args, w=None, **kwargs):
    
    if not isMLSWeightFunction(w):
        dim = points.shape[1]
        w = ConstantWeightFunction(dim)
    
    def inner(x):
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise TypeError
        w.core = x
        f = weighted_least_squares(points, *args, w=w, **kwargs)
        return f(x)

    return inner


def least_squares(*args, **kwargs):
    return weighted_least_squares(*args, **kwargs)


def weighted_least_squares(points, values, *args, deg=1, order=2, w=None, **kwargs):
    """
    Returns a Callable that can be used to approximate over datasets.

    Parameters
    ----------
    points : Iterable 
        [[X11, X12, ..., X1d], ..., [Xn1, Xn2, ..., Xnd]]

    values : Iterable
        [[f11, f12, ..., f1r], ..., [fn1, fn2, ..., fnr]]

    deg : int, Optional
        The degree of the approximation. Default is 1.

    dim : int, Optional
        The dimension of the dataset. Default is 1.

    n : int, Optional
        Number of data points per dimension. Default is 20.

    w : MLSWeightFunction, Optional
        A proper weight function. Default is a ConstantWeightFunction.

    order : int, Optional.
        The order of the approximation. Default is 2.

    Returns
    -------
    Callable
        Regression function r(x) = f(x), fdx(x), fdy(x), fdxx(x), fdyy(x), fdxy(x)
        fi([X1, X2, ..., Xd]) = [fi1, fi2,..., fir]

    Note
    ----
    The resulting approximation can have an approximation or regression behaviour,
    depending on the data set and the degree of the polynomial.

    """

    assert isinstance(points, np.ndarray)
    assert isinstance(values, np.ndarray)
    assert points.shape[0] == values.shape[0]
    dim = points.shape[1]
    try:
        rec = values.shape[1]
    except:
        rec = 1

    if isMLSWeightFunction(w):
        assert dim == w.dimension
    else:
        w = ConstantWeightFunction(dim)

    grad = True if order > 0 else False
    hess = True if order > 1 else False
    if grad:
        assert hasattr(w, 'gradient')        
    if hess:
        grad = True
        assert hasattr(w, 'gradient')
        assert hasattr(w, 'Hessian')

    def bdx(x): return None
    def bdy(x): return None
    def bdxx(x): return None
    def bdyy(x): return None
    def bdxy(x): return None
    if deg == 1:
        if dim == 1:
            def b(x): return np.array([1, x])
            if grad:
                def bdx(x): return np.array([0, 1])
            if hess:
                def bdxx(x): return np.array([0, 0])
        elif dim == 2:
            def b(x): return np.array([1, x[0], x[1]])
            if grad:
                def bdx(x): return np.array([0, 1, 0])
                def bdy(x): return np.array([0, 0, 1])
            if hess:
                def bdxx(x): return np.array([0, 0, 0])
                def bdyy(x): return np.array([0, 0, 0])
                def bdxy(x): return np.array([0, 0, 0])
        else:
            raise
    elif deg == 2:
        if dim == 1:
            def b(x): return np.array([1, x, x**2])
            if grad:
                def bdx(x): return np.array([0, 1, 2*x[0]])
            if hess:
                def bdxx(x): return np.array([0, 0, 2])
        elif dim == 2:
            def b(x): return np.array(
                [1, x[0], x[1], x[0]**2, x[1]**2, x[0]*x[1]])
            if grad:
                def bdx(x): return np.array([0, 1, 0, 2*x[0], 0, x[1]])
                def bdy(x): return np.array([0, 0, 1, 0, 2*x[1], x[0]])
            if hess:
                def bdxx(x): return np.array([0, 0, 0, 2, 0, 0])
                def bdyy(x): return np.array([0, 0, 0, 0, 2, 0])
                def bdxy(x): return np.array([0, 0, 0, 0, 0, 1])
        else:
            raise
    else:
        raise

    # moment matrix
    nData = points.shape[0]
    k = int(fact(deg+dim)/fact(deg)/fact(dim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Ady, Bdx, Bdy = 4 * (None,)
    Adxx, Adyy, Adxy, Bdxx, Bdyy, Bdxy = 6 * (None,)
    if grad:
        Adx = np.zeros([k, k])
        Ady = np.zeros([k, k])
        Bdx = np.zeros([k, nData])
        Bdy = np.zeros([k, nData])
    if hess:
        Adxx = np.zeros([k, k])
        Adyy = np.zeros([k, k])
        Adxy = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])
        Bdyy = np.zeros([k, nData])
        Bdxy = np.zeros([k, nData])

    V = np.zeros([nData, rec])
    for i, (xi, fi) in enumerate(zip(points, values)):
        bi = b(xi)
        wi = w.f(xi)
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if grad:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Ady += Mi * gwi[1]
            Bdx[:, i] += bi * gwi[0]
            Bdy[:, i] += bi * gwi[1]
        if hess:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Adyy += Mi * Gwi[1, 1]
            Adxy += Mi * Gwi[0, 1]
            Bdxx[:, i] += bi * Gwi[0, 0]
            Bdyy[:, i] += bi * Gwi[1, 1]
            Bdxy[:, i] += bi * Gwi[0, 1]
        for r in range(rec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    def inner(x):
        return mls_approx(invA, B, b(x), V,
                          Adx, Ady, Adxx, Adyy, Adxy,
                          Bdx, Bdy, Bdxx, Bdyy, Bdxy,
                          bdx(x), bdy(x), bdxx(x), bdyy(x), bdxy(x),
                          g=grad, H=hess)
    return inner


def mls_preproc(points, values, deg, w, b, g=True, H=True):
    nData = points.shape[0]
    nDim = points.shape[1]
    nRec = values.shape[1]

    # moment matrix
    k = int(fact(deg+nDim)/fact(deg)/fact(nDim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Ady, Bdx, Bdy = None, None, None, None
    Adxx, Adyy, Adxy, Bdxx, Bdyy, Bdxy = None, None, None, None, None, None
    if g:
        Adx = np.zeros([k, k])
        Ady = np.zeros([k, k])
        Bdx = np.zeros([k, nData])
        Bdy = np.zeros([k, nData])
    if H:
        Adxx = np.zeros([k, k])
        Adyy = np.zeros([k, k])
        Adxy = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])
        Bdyy = np.zeros([k, nData])
        Bdxy = np.zeros([k, nData])

    V = np.zeros([nData, nRec])
    for i in range(nData):
        xi = points[i]
        fi = values[i]
        bi = b[i]
        wi = w[i]
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if g:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Ady += Mi * gwi[1]
            Bdx[:, i] += bi * gwi[0]
            Bdy[:, i] += bi * gwi[1]
        if H:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Adyy += Mi * Gwi[1, 1]
            Adxy += Mi * Gwi[0, 1]
            Bdxx[:, i] += bi * Gwi[0, 0]
            Bdyy[:, i] += bi * Gwi[1, 1]
            Bdxy[:, i] += bi * Gwi[0, 1]
        for r in range(nRec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    return invA, V, B, Adx, Ady, Adxx, Adyy, Adxy, Bdx, Bdy, Bdxx, Bdyy, Bdxy


def mls_approx(invA, B, b, V,
               Adx, Ady, Adxx, Adyy, Adxy,
               Bdx, Bdy, Bdxx, Bdyy, Bdxy,
               bdx, bdy, bdxx, bdyy, bdxy,
               g=True, H=True):
    gamma = invA @ b
    SHP = gamma @ B
    f = SHP.T @ V
    gammadx = invA @ (bdx - Adx @ gamma)
    gammady = invA @ (bdy - Ady @ gamma)
    fdxx = None
    fdyy = None
    fdxy = None
    if g:
        fdx, fdy = mls_g(gamma, gammadx, gammady, B, Bdx, Bdy, V)
    if H:
        fdxx, fdyy, fdxy = mls_H(invA, bdxx, bdyy, bdxy, Adx, Ady,
                                 Adxx, Adyy, Adxy, gamma, gammadx, gammady,
                                 B, Bdx, Bdy, Bdxx, Bdyy, Bdxy, V)
    return f, fdx, fdy, fdxx, fdyy, fdxy


@njit(nogil=True, parallel=True, fastmath=True, cache=True)
def mls_g(gamma, gammadx, gammady, B, Bdx, Bdy, V):
    SHPdx = gammadx @ B + gamma @ Bdx
    SHPdy = gammady @ B + gamma @ Bdy
    fdx = SHPdx.T @ V
    fdy = SHPdy.T @ V
    return fdx, fdy


@njit(nogil=True, parallel=True, fastmath=True, cache=True)
def mls_H(invA, bdxx, bdyy, bdxy, Adx, Ady, Adxx, Adyy, Adxy,
          gamma, gammadx, gammady, B, Bdx, Bdy, Bdxx, Bdyy, Bdxy, V):
    gammadxx = invA @ (bdxx - Adxx @ gamma - 2 * Adx @ gammadx)
    gammadyy = invA @ (bdyy - Adyy @ gamma - 2 * Ady @ gammady)
    gammadxy = invA @ (bdxy - Adxy @ gamma - Adx @ gammady - Ady @ gammadx)
    SHPdxx = gammadxx @ B + 2 * gammadx @ Bdx + gamma @ Bdxx
    SHPdyy = gammadyy @ B + 2 * gammady @ Bdy + gamma @ Bdyy
    SHPdxy = gammadxy @ B + gammadx @ Bdy + gammady @ Bdx + gamma @ Bdxy
    fdxx = SHPdxx.T @ V
    fdyy = SHPdyy.T @ V
    fdxy = SHPdxy.T @ V
    return fdxx, fdyy, fdxy



# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np


__all__ = ['histogram']


def histogram(data, bins, *args, return_edges=False, **kwargs):
    hist, bin_edges = np.histogram(data, bins, *args, **kwargs)
    if return_edges:
        return hist, bin_edges
    else:
        bc = bin_centers(bin_edges)
        return hist, bc


@njit(nogil=True, parallel=True, cache=True)
def bin_centers(bin_edges: np.ndarray):
    res = np.zeros(bin_edges.shape[0]-1, dtype=bin_edges.dtype)
    for i in prange(res.shape[0]):
        res[i] = (bin_edges[i] + bin_edges[i + 1]) / 2
    return res



# -*- coding: utf-8 -*-
from typing import Callable
from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True

__all__ = ['to_range_1d', 'squeeze_if_array',
           'config', 'is_none_or_false', 'squeeze']


@njit(nogil=True, parallel=True, cache=__cache)
def _to_range_1d_(vals: ndarray, source: ndarray, target: ndarray):
    res = np.zeros_like(vals)
    s0, s1 = source
    t0, t1 = target
    b = (t1 - t0) / (s1 - s0)
    a = (t0 + t1) / 2 - b * (s0 + s1) / 2
    for i in prange(res.shape[0]):
        res[i] = a + b * vals[i]
    return res


def to_range_1d(vals: ndarray, *args, source: ndarray, target: ndarray = None,
                squeeze=False, **kwargs):
    if not isinstance(vals, ndarray):
        vals = np.array([vals, ])
    source = np.array([0., 1.]) if source is None else np.array(source)
    target = np.array([-1., 1.]) if target is None else np.array(target)
    if squeeze:
        return np.squeeze(_to_range_1d_(vals, source, target))
    else:
        return _to_range_1d_(vals, source, target)


def squeeze_if_array(arr):
    return np.squeeze(arr) if isinstance(arr, np.ndarray) else arr


def squeeze(default=True):
    def decorator(fnc: Callable):
        def inner(*args, squeeze:bool=default, **kwargs):
            if squeeze:
                res = fnc(*args, **kwargs)
                if isinstance(res, tuple):
                    return list(map(squeeze_if_array, res))
                elif isinstance(res, dict):
                    return {k : squeeze_if_array(v) for k, v in res.items()}
                else:
                    return squeeze_if_array(res)
            else:
                return fnc(*args, **kwargs)
        inner.__doc__ = fnc.__doc__
        return inner
    return decorator


def config(*args, **kwargs):
    def decorator(fnc: Callable):
        def inner(*args, **kwargs):
            return fnc(*args, **kwargs)
        return inner
    return decorator


def is_none_or_false(a):
    if isinstance(a, bool):
        return not a
    elif a is None:
        return True
    return False



# -*- coding: utf-8 -*-
import collections

from .function import Function



class TestFunction(Function):

    __slots__ = ('optimums', 'optText')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optimums = []

        self.optText = 'opt'
        for key, value in kwargs.items():
            if key == 'optText':
                assert isinstance(value, str)
                self.optText = value
                break
            elif key == 'optimums':
                assert isinstance(value, collections.Iterable)
                for v in value:
                    self.optimums.append(v)
        return


class TestFunction2D(TestFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, d=2)


class TestMinFunction(TestFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, optText='min')


class TestMaxFunction(TestFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, optText='max')


class TestMinFunction2D(TestMinFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, d=2)


class TestMaxFunction2D(TestMaxFunction):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, d=2)



# -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import LinAlgError

from .proc import frontal_sym_bulk_uniform
from .postproc import backsub_fr
from .path import order_to_path


def frsolve(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray = None,
            presc_val: np.ndarray = None, topology: np.ndarray = None,
            epath: np.ndarray = None, eorder=None):
    try:
        nEQ = len(B)
        if len(B.shape) == 1:
            B = B.reshape((nEQ, 1))
        if epath is None:
            if eorder is not None:
                epath = order_to_path(eorder)
            else:
                epath = np.arange(len(A))
        pre = presc_val is not None
        if not pre:
            presc_bool = np.zeros((nEQ,), dtype=int)
            presc_val = np.zeros((nEQ,), dtype=float)
        lhs, rhs, eqpath, glob_to_front, glob_to_width = \
            frontal_sym_bulk_uniform(A, topology, B,
                                     presc_bool, presc_val, epath)
        res = backsub_fr(lhs, rhs, presc_bool, presc_val, eqpath,
                         glob_to_front, glob_to_width)
        if pre:
            return res
        else:
            return res[0]
    except Exception:
        raise LinAlgError('The matrix is singular!')


if __name__ == '__main__':
    pass
    """
    from polydata.mechanics.fem import Uniform3D
    from polydata.mechanics.material.hooke import Hooke
    from dewloosh.math.linalg import solve
    from polydata.core import NestedDefaultDict
    import numpy as np
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import spsolve
    from time import time

    solutions = NestedDefaultDict()
    solvers = ['numpy']
    for solver in solvers:

        # geometry
        Lx, Ly, Lz = shape = (100, 100, 100)

        # mesh control
        nx, ny, nz = size = 3 * (11,)

        # assembly
        assembly = Uniform3D(shape, size)
        cells = assembly.cells()
        points = assembly.points()
        coord = np.array(points)

        # material
        model = Hooke(E=2000, NU=0.3)
        for c in cells:
            c.model = model

        # essential boundary conditions
        condZ = (abs(coord[:, 2] - Lz/2) < 0.001)
        condX = (coord[:, 0] < 0.001) | (coord[:, 0] > Lx-0.001)
        condY = (coord[:, 1] < 0.001) | (coord[:, 1] > Ly-0.001)
        cond = (condX | condY) & condZ
        nIDs = np.where(cond)[0]
        if solver == 'numpy':
            p = np.array([1e12, 1e12, 1e12], dtype=np.float64)
            penalties = {nID: p for nID in nIDs}
        else:
            disps = {}
            for nID in nIDs:
                for idof in range(3):
                    disps[nID * 3 + idof] = 0.0

        # natural boundary conditions
        case = 0
        if case == 0:
            nIDs = np.where(condZ)[0]
            fz = (-10/10000)*Lx*Ly/len(nIDs)
            fp = np.array([0, 0, fz], dtype=np.float64)
            loads = {nID: fp for nID in nIDs}
        elif case == 1:
            condZ = (abs(coord[:, 2] - Lz/2) < 0.001)
            condX = (abs(coord[:, 0] - Lx/2) < 0.001)
            condY = (abs(coord[:, 1] - Ly/2) < 0.001)
            cond = condX & condY & condZ
            nIDs = np.where(cond)[0]
            loads = {nIDs[0]: np.array([0., 0., -10.0])}

        # preprocessor
        if solver == 'frontal':
            assembly.preprocess(solver=solver)
        elif solver == 'numpy':
            assembly.preprocess(solver=solver)
        elif solver == 'Gauss-Jordan':
            assembly.preprocess(solver='numpy')
        elif solver == 'Jordan':
            assembly.preprocess(solver='numpy')

        # solve
        nTOTV = len(points) * 3
        if solver == 'frontal':
            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            K, Ktopo = assembly.stiffness_matrix_frontal()

            # connectivity and topology
            adj_csr = assembly.adjacency_matrix(sparse=True, mode='node')
            topo_csr = assembly.connectivity_matrix(sparse=True,
                                                    mode='node')
            epath = np.arange(len(cells))

            # solution
            A = K
            B = RHS.reshape((nTOTV, 1))
            topology = Ktopo
            t1 = time()
            U, R = frsolve(A, B, presc_bool, presc_val, topology, epath)
            t2 = time()
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)
        elif solver == 'numpy':
            K = assembly._K

            # essential boundary conditions
            spind = np.concatenate([points[ID].gdofs
                                    for ID in penalties.keys()])
            spdata = np.concatenate([v for v in penalties.values()])
            K += coo_matrix((spdata, (spind, spind)), shape=(nTOTV, nTOTV),
                            dtype=np.float64)

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # solution
            CSR = K.tocsr()
            f = RHS.flatten()
            t1 = time()
            u = spsolve(CSR, f)
            t2 = time()
            assembly.point_data['dof'] = u.reshape(RHS.shape)
            assembly.point_data['res'] = \
                np.reshape(f - np.matmul(K.todense(), u), RHS.shape)
        elif solver == 'Gauss-Jordan':
            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            A = assembly._K.todense()
            B = RHS.reshape((nTOTV, 1))
            t1 = time()
            U, R = solve(A, B, presc_bool, presc_val, method='Gauss-Jordan')
            t2 = time()
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)
        elif solver == 'Jordan':
            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            A = assembly._K.todense()
            B = RHS.reshape((nTOTV, 1))
            t1 = time()
            U, R = solve(A, B, presc_bool, presc_val, method='Jordan')
            t2 = time()
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)

        solutions[solver]['umin'] = np.min(assembly.point_data['dof'])
        solutions[solver]['umax'] = np.max(assembly.point_data['dof'])
        solutions[solver]['u[0]'] = assembly.point_data['dof'][0]
        solutions[solver]['time'] = (t2 - t1)

    print(solutions['frontal']['umax'])
    print(solutions['numpy']['umax'])
    print(solutions['Gauss-Jordan']['umax'])
    print('\n')
    print(solutions['frontal']['time'])
    print(solutions['numpy']['time'])
    print(solutions['Gauss-Jordan']['time'])
    """


# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange, prange

__cache = True


@njit(nogil=True, cache=__cache)
def show_vector(dcm: np.ndarray, arr: np.ndarray):
    """
    Returns the coordinates of a single vector in a frame specified
    by a DCM matrix.
    
    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    
    arr : numpy.ndarray
        1d float array of coordinates of a single vector.
    
    Returns
    -------      
    numpy.ndarray
        The new coordinates of the vector with the same shape as `arr`.
        
    """
    return dcm @ arr


@njit(nogil=True, parallel=True, cache=__cache)
def show_vectors(dcm: np.ndarray, arr: np.ndarray):
    """
    Returns the coordinates of multiple vectors in a frame specified
    by a DCM matrix.
    
    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 2d float array.
    
    arr : numpy.ndarray
        2d float array of coordinates of multiple vectors.
    
    Returns
    -------      
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
        
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm @ arr[i, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def show_vectors_multi(dcm: np.ndarray, arr: np.ndarray):
    """
    Returns the coordinates of multiple vectors and multiple DCM matrices.
    
    Parameters
    ----------
    dcm : numpy.ndarray
        The dcm matrix of the transformation as a 3d float array.
    
    arr : numpy.ndarray
        2d float array of coordinates of multiple vectors.
    
    Returns
    -------      
    numpy.ndarray
        The new coordinates of the vectors with the same shape as `arr`.
        
    """
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm[i] @ arr[i, :]
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from awkward import Array as akarray
from numba import njit
from numba.core import types
from numba.typed import Dict
from typing import Union
import awkward as ak

from .array import flatten2dC
from .array import count_cols

__cache = True


ArrayLike = Union[ndarray, akarray]
i64 = types.int64
i64A = types.int64[:]
i64A2 = types.int64[:, :]


def unique2d(arr: ArrayLike, return_index=False,
             return_inverse=False, return_counts=False):
    if isinstance(arr, ndarray):
        unique, counts, inverse, indices = _unique2d_njit(arr)
        res = [unique, ]
        if return_index:
            res.append(indices)
        if return_inverse:
            res.append(inverse)
        if return_counts:
            res.append(counts)
        return res if len(res) > 1 else res[0]
    elif isinstance(arr, akarray):
        assert arr.ndim == 2, "Only 2 dimensional awkward arrays are supported!"
        if hasattr(arr, 'is_jagged'):
            if not arr.is_jagged():
                return unique2d(arr.to_numpy(), return_index=return_index,
                                return_inverse=return_inverse,
                                return_counts=return_counts)
        flatarr = ak.flatten(arr).to_numpy()
        unique, counts, inverse, indices = _unique1d_njit(flatarr)
        res = [unique, ]
        if return_index:
            res.append(indices)
        if return_inverse:
            res.append(ak.unflatten(inverse, count_cols(arr)))
        if return_counts:
            res.append(counts)
        return res if len(res) > 1 else res[0]


@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def _unique1d_njit(flatdata: ndarray):
    unique = np.unique(flatdata)
    nF = len(flatdata)
    nU = len(unique)
    counts = np.zeros(nU, dtype=np.int64)
    inverse = np.zeros(nF, dtype=np.int64)
    imap = Dict.empty(key_type=i64, value_type=i64)
    for i in range(nU):
        imap[unique[i]] = i
    for i in range(nF):
        counts[imap[flatdata[i]]] += 1
        inverse[i] = imap[flatdata[i]]
    indices = Dict.empty(key_type=i64, value_type=i64A)
    for i in range(nU):
        indices[unique[i]] = np.zeros((counts[i]), dtype=np.int64)
    counts[:] = 0
    for iF in range(nF):
        i = inverse[iF]
        indices[unique[i]][counts[i]] = iF
        counts[i] += 1
    return unique, counts, inverse, indices


@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def _unique2d_njit(data: ndarray):
    flatdata = flatten2dC(data)
    unique = np.unique(flatdata)
    nF = len(flatdata)
    nU = len(unique)
    counts = np.zeros(nU, dtype=np.int64)
    inverse = np.zeros(nF, dtype=np.int64)
    imap = Dict.empty(key_type=i64, value_type=i64)
    for i in range(nU):
        imap[unique[i]] = i
    for i in range(nF):
        counts[imap[flatdata[i]]] += 1
        inverse[i] = imap[flatdata[i]]
    inverse = inverse.reshape(data.shape)
    indices = Dict.empty(key_type=i64, value_type=i64A2)
    for i in range(nU):
        indices[unique[i]] = np.zeros((counts[i], 2), dtype=np.int64)
    counts[:] = 0
    nR, nC = data.shape
    for r in range(nR):
        for c in range(nC):
            i = inverse[r, c]
            indices[unique[i]][counts[i], 0] = r
            indices[unique[i]][counts[i], 1] = c
            counts[i] += 1
    return unique, counts, inverse, indices

"""
@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def unique2d_njit_jagged(data: akarray):
    return ak.flatten(data)
"""


# -*- coding: utf-8 -*-
from typing import TypeVar, Callable
import sympy as sy
from sympy import Expr, degree, latex, lambdify
from sympy.core.numbers import One
import numpy as np
from collections import OrderedDict

from dewloosh.core.tools import getasany

from .metafunction import MetaFunction, substitute


__all__ = ['Function', 'VariableManager', 'FuncionLike']


FuncionLike = TypeVar('FuncionLike', str, Callable, Expr)


class Function(MetaFunction):
    """
    Base class for all kinds of functions.

    Parameters
    ----------
    f0 : Callable
        A callable object that returns function evaluations.

    f1 : Callable
        A callable object that returns evaluations of the 
        gradient of the function.

    f2 : Callable
        A callable object that returns evaluations of the 
        Hessian of the function.

    variables : List, Optional
        Symbolic variables. Only if the function is defined by 
        a string or `sympy` expression.

    value : Callable, Optional
        Same as `f0`.

    gradient : Callable, Optional
        Same as `f1`.

    Hessian : Callable, Optional
        Same as `f2`.

    dimension or dim or d : int, Optional
        The number of dimensions of the domain of the function. Required only when
        going full blind, in most of the cases it can be derived from other properties.

    Examples
    --------
    >>> from neumann.function import Function
    >>> import sympy as sy
    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> f = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
    >>> f([0, 6, 0, 4])
    10

    >>> def f0(x=None, y=None):
    >>>     return x**2 + y

    >>> def f1(x=None, y=None):
    >>>     return np.array([2*x, 1])

    >>> def f2(x=None, y=None):
    >>>     return np.array([[0, 0], [0, 0]])

    >>> f = Function(f0, f1, f2, d=2)
    >>> f.linear
    ...

    >>> g = Function('3*x + 4*y - 2', variables=['x', 'y', 'z'])
    >>> g.linear
    ...

    >>> h = Function('3*x + 4*y - 2')
    >>> h.linear
    ...

    >>> m = Function('3*x + 4*y - 2', variables=symbols('x y z'))
    >>> m.linear
    ...

    >>> m([1, 2, -30])
    ...

    """
    # FIXME domain is missing from the possible parameters
    # NOTE investigate if dimensions should be derived

    def __init__(self, f0: FuncionLike = None, f1: Callable = None,
                 f2: Callable = None, *args, variables=None, **kwargs):
        super().__init__()
        self.update(f0, f1, f2, *args, variables=variables, **kwargs)

    def update(self, f0: FuncionLike = None, f1: Callable = None,
               f2: Callable = None, *args, variables=None, **kwargs):
        self.from_str = None
        if f0 is not None:
            if isinstance(f0, str):
                kwargs.update(self._str_to_func(
                    f0, variables=variables, **kwargs))
                self.from_str = True
            elif isinstance(f0, Expr):
                kwargs.update(self._sympy_to_func(
                    f0, variables=variables, **kwargs))
        self.expr = kwargs.get('expr', None)
        self.variables = kwargs.get('variables', variables)
        self.f0 = kwargs.get('value', f0)
        self.f1 = kwargs.get('gradient', f1)
        self.f2 = kwargs.get('Hessian', f2)
        self.dimension = getasany(['d', 'dimension', 'dim'], None, **kwargs)
        self.domain = kwargs.get('domain', None)
        self.vmap = kwargs.get('vmap', None)

    @property
    def symbolic(self):
        """
        Returns True if the function is a fit subject of symbolic manipulation.
        This is probably only true if the object was created from a string or
        `sympy` expression.
        """
        try:
            return self.expr is not None
        except AttributeError:
            return False

    @property
    def linear(self):
        """
        Returns True if the function is at most linear in all of its variables.
        """
        if self.symbolic:
            return all(np.array([degree(self.expr, v)
                                 for v in self.variables], dtype=int) <= 1)
        else:
            return self.f2 is None

    def linear_coefficients(self, normalize=False):
        d = self.coefficients(normalize)
        if d:
            return {key: value for key, value in d.items()
                    if len(key.free_symbols) <= 1}
        return None

    def coefficients(self, normalize=False):
        try:
            d = OrderedDict({x: 0 for x in self.variables})
            d.update(self.expr.as_coefficients_dict())
            if not normalize:
                return d
            else:
                res = OrderedDict()
                for key, value in d.items():
                    if len(key.free_symbols) == 0:
                        res[One()] = value*key
                    else:
                        res[key] = value
                return res
        except Exception:
            return None

    def to_latex(self):
        """
        Returns the LaTeX code of the symbolic expression of the object.

        Only for simbolic functions.

        """
        assert self.symbolic, "This is exclusive to symbolic functions."
        try:
            return latex(self.expr)
        except Exception:
            return None

    def subs(self, values, variables=None, inplace=False):
        """
        Substitites values for variables.
        
        """
        assert self.symbolic, "This is exclusive to symbolic functions."
        if self.expr is None:
            return None
        expr = substitute(self.expr, values, variables,
                          as_string=self.from_str)
        kwargs = self._sympy_to_func(expr=expr, variables=variables)
        if not inplace:
            return Function(None, None, None, **kwargs)
        else:
            self.update(None, None, None, **kwargs)
            return self


class VariableManager(object):

    def __init__(self, variables=None, vmap=None, **kwargs):
        try:
            variables = list(sy.symbols(variables, **kwargs))
        except Exception:
            variables = variables
        try:
            self.vmap = vmap if vmap is not None else OrderedDict(
                {v: v for v in variables})
        except Exception:
            self.vmap = OrderedDict()
        self.variables = variables  # this may be unnecessary

    def substitute(self, vmap: dict = None, inverse=False, inplace=True):
        if not inverse:
            sval = list(vmap.values())
            svar = list(vmap.keys())
        else:
            sval = list(vmap.keys())
            svar = list(vmap.values())
        if inplace:
            for v, expr in self.vmap.items():
                self.vmap[v] = substitute(expr, sval, svar)
            return self
        else:
            vmap = OrderedDict()
            for v, expr in self.vmap.items():
                vmap[v] = substitute(expr, sval, svar)
            return vmap

    def lambdify(self, variables=None):
        assert variables is not None
        for v, expr in self.vmap.items():
            self.vmap[v] = lambdify([variables], expr, 'numpy')

    def __call__(self, v):
        return self.vmap[v]

    def target(self):
        return list(self.vmap.keys())

    def source(self):
        s = set()
        for expr in self.vmap.values():
            s.update(expr.free_symbols)
        return list(s)

    def add_variables(self, variables, overwrite=True):
        if overwrite:
            self.vmap.update({v: v for v in variables})
        else:
            for v in variables:
                if v not in self.vmap:
                    self.vmap[v] = v



# -*- coding: utf-8 -*-
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


def alphabet(abctype: str = 'latin', **kwargs) -> Iterable:
    if abctype in ('ord', 'o'):
        start = kwargs.pop('start', 0)
    elif abctype in ('latin', 'l'):
        start = ord(kwargs.pop('start', 'a'))
    elif abctype == 'u':
        start = ord(kwargs.pop('start', '\u0000'))
    elif abctype in ('greek', 'g'):
        start = ord(kwargs.pop('start', '\u03b1'))
    while True:
        yield chr(start)
        start += 1


def ordrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop('start', 0)
    if isinstance(start, str):
        start = ord(start)
    stop = kwargs.pop('stop', None)
    if stop is None or stop == start:
        stop = start + N
    return [chr(c) for c in range(start, stop)]


def latinrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop('start', 97)
    stop = kwargs.pop('stop', None)
    return ordrange(N, start=start, stop=stop)


def urange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop('start', '\u0000')
    stop = kwargs.pop('stop', None)
    if stop is None:
        stop = start
    return ordrange(N, start=ord(start), stop=ord(stop))


def greekrange(N: int = 1) -> Iterable:
    return urange(N, start='\u03b1')


def arabicrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop('start', 0)
    stop = kwargs.pop('stop', None)
    if stop is None or stop == start:
        stop = start + N
    return [str(c) for c in range(start, stop)]


from typing import Union
import numpy as np
from numpy import ndarray
from .array import atleast1d


def isfloatarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a float array.
    """
    return np.issubdtype(a.dtype, float)


def isintegerarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is an integer array.
    """
    return np.issubdtype(a.dtype, int)


def isintarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a integer array.
    """
    return isintegerarray(a)


def isboolarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a boolean array.
    """
    return np.issubdtype(a.dtype, bool)


def is1dfloatarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a 1d float array.
    """
    return isfloatarray(a) and len(a.shape) == 1


def is1dintarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a 1d integer array.
    """
    return isintarray(a) and len(a.shape) == 1


def issymmetric(a: ndarray, tol:float=1e-8) -> bool:
    """
    Returns `True` if `a` is symmetric with a given tolerance
    prescribed by `tol`. 
    """
    return np.linalg.norm(a-a.T) < tol


def isposdef(A: ndarray, tol=0) -> bool:
    """
    Returns `True` if `A` is positive definite.

    Example
    -------
    >>> from neumann.array import random_posdef_matrix
    >>> from neumann.logical import isposdef
    >>> A = random_posdef_matrix(3, 0.1)
    >>> isposdef(A)
    True
    
    >>> A[0, 0] = 0
    >>> isposdef(A)
    False
    
    """
    return np.all(np.linalg.eigvals(A) > tol)


def ispossemidef(A: ndarray) -> bool:
    """
    Returns `True` if `A` is positive semidefinite.

    Example
    -------
    >>> from neumann.array import random_pos_semidef_matrix
    >>> from neumann.logical import ispossemidef
    >>> A = random_pos_semidef_matrix(3)
    >>> ispossemidef(A)
    True
    """
    return np.all(np.linalg.eigvals(A) >= 0)


def isclose(x1: ndarray, x2: ndarray, *, atol:float=1e-8, 
            rtol:float=1e-5) -> Union[bool, ndarray]:
    """
    Returns a boolean array where two arrays are element-wise equal 
    in absolute and optionally in relative sense. In the latter case
    the relative difference is measured agains `x2`.
    
    Absolute difference is measured as abs(`x1` - `x2`) <= `atol`, relative
    difference is measured as abs(`x1` - `x2`) <= abs(`rtol` * `x2`).
    
    .. versionadded:: 0.0.6
    
    .. warning:: Default values might fail for very small numbers.
    
    Parameters
    ----------
    x1 : ndarray or float
        Values of the first array.
        
    x2 : ndarray or float
        Values of the second array.
        
    atol : float, Optional
        Absolute tolerance. It must be a positive number, or None to
        turn this check off. Default is 1e-8.
        
    rtol : float, Optional
        Relative tolerance. It must be a positive number or None to
        turn this check off. In the latter case, the input arrays are only
        compared in an absolute sense. Default is 1e-5.
    
    Returns
    -------
    numpy.ndarray or float
        A boolean array or a float, depending on the input.
        
    See Also
    --------
    numpy.isclose
    numpy.allclose
    math.isclose    
    """
    if not isinstance(x1, ndarray):
        x1 = atleast1d(x1)
    if not isinstance(x2, ndarray):
        x2 = atleast1d(x2)
    c = np.ones(x1.shape, dtype=bool)
    assert (atol is not None) or (rtol is not None), \
        "At least one from `atol` or `rtol` must be other than `None`."
    if atol is not None:
        assert atol > 0, "The absolute tolerance must be a positive number."
        c_abs = (np.abs(x1 - x2) <= atol)
        c = c & c_abs
    if rtol is not None:
        assert rtol > 0, "The relative tolerance must be a positive number."
        c_rel = (np.abs(x1 - x2) <= np.abs(rtol * x2))
        c = c & c_rel
    if len(c.shape) == 1 and c.shape[0] == 1:
        return c[0]
    return c


def allclose(*args, **kwargs) -> bool:
    """
    Same as `isclose`, but it returns a single boolean that is True
    ir False if all the values returned by `isclose` are True or False.
    
    See Also
    --------
    isclose
    """
    return np.all(isclose(*args, **kwargs))
    




# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange
from typing import Union, Tuple
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
__cache = True

ArrayOrFloat = Union[float, ndarray, list]


def itype_of_ftype(dtype):
    """
    Returns a matching NumPy integer type to a float type.
    """
    name = np.dtype(dtype).name
    if '32' in name:
        return np.int32
    elif '64' in name:
        return np.int64
    else:
        raise TypeError('Unrecognized float type.')
    

@njit(nogil=True, cache=__cache)
def minmax(a: ndarray) -> Tuple[float]:
    """
    Returns the minimum and maximum values of an array.
    """
    return a.min(), a.max()


def ascont(array: ndarray) -> ndarray:
    """
    Returns the input as contiguous array.
    It is basically a shortcut to `numpy.ascontiguousarray`.
    """
    return np.ascontiguousarray(array)


@njit(nogil=True, cache=__cache)
def clip1d(a: ndarray, a_min: float, a_max: float) -> ndarray:
    """
    Clips the values outside the interval [a_min, a_max] to 
    either a_min or a_max.
    """
    a[a < a_min] = a_min
    a[a > a_max] = a_max
    return a


def atleastnd(a: ndarray, n:int=2, front:bool=True, 
              back:bool=False) -> ndarray:
    """
    Returns an array that is at least 'n' dimensional.
    The required shape is obtained by inserting new axes either
    before or after existing ones. This behaviour can be controlled
    using the parameters 'front' and 'back'. If front is True and back 
    is False, new axes are crated before the first existing data index,
    and the opposite happens in every other case.
    """
    shp = a.shape
    nD = len(shp)
    if nD >= n:
        return a
    else:
        if front and not back:
            newshape = (n - nD) * (1,) + shp
        else:
            newshape = shp + (n - nD) * (1,)
        return np.reshape(a, newshape)


def atleast1d(a: ArrayOrFloat) -> ndarray:
    """Returns an array that is at least 1 dimensional."""
    if not isinstance(a, Iterable):
        a = [a,]
    return np.array(a)


def atleast2d(a: ndarray, **kwargs) -> ndarray:
    """Returns an array that is at least 2 dimensional."""
    return atleastnd(a, 2, **kwargs)


def matrixform(f: ndarray) -> ndarray:
    """Returns an array that is at least 2 dimensional."""
    size = len(f.shape)
    assert size <= 2, "Input array must be at most 2 dimensional."
    if size == 1:
        nV, nC = len(f), 1
        return f.reshape(nV, nC)
    return f


def atleast3d(a: ndarray, **kwargs) -> ndarray:
    """Returns an array that is at least 3 dimensional."""
    return atleastnd(a, 3, **kwargs)


def atleast4d(a: ndarray, **kwargs) -> ndarray:
    """Returns an array that is at least 4 dimensional."""
    return atleastnd(a, 4, **kwargs)


@njit(nogil=True, cache=__cache)
def flatten2dC(a: ndarray) -> ndarray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for i in range(I):
        for j in range(J):
            res[ind] = a[i, j]
            ind += 1
    return res


@njit(nogil=True, cache=__cache)
def flatten2dF(a: ndarray) -> ndarray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for j in range(J):
        for i in range(I):
            res[ind] = a[i, j]
            ind += 1
    return res


def flatten2d(a: ndarray, order: str = 'C') -> ndarray:
    """
    Returns a flattened view of `a`.
    """
    if order == 'C':
        return flatten2dC(a)
    elif order == 'F':
        return flatten2dF(a)


def bool_to_float(a: ndarray, true=1.0, false=0.0) -> ndarray:
    """
    Transforms a boolean array to a float array using the specified
    values for `True` and `False`. 
    """
    res = np.full(a.shape, false, dtype=float)
    res[a] = true
    return res


def choice(choices, size, probs=None) -> ndarray:
    """
    Returns a numpy array, whose elements are selected from
    'choices' under probabilities provided with 'probs' (optionally).

    Example
    -------
    >>> N, p = 10, 0.2
    >>> choice([False, True], (N, N), [p, 1-p])
    
    """
    if probs is None:
        probs = np.full((len(choices),), 1/len(choices))
    return np.random.choice(a=choices, size=size, p=probs)


def random_pos_semidef_matrix(N) -> ndarray:
    """
    Returns a random positive semidefinite matrix of shape (N, N).

    Example
    -------
    >>> from neumann.array import random_pos_semidef_matrix
    >>> random_pos_semidef_matrix(3)
    ...
    """
    A = np.random.rand(N, N)
    return A.T @ A


def random_posdef_matrix(N, alpha=1e-12) -> ndarray:
    """
    Returns a random positive definite matrix of shape (N, N).
    
    All eigenvalues of this matrix are >= alpha.

    Example
    -------
    >>> from neumann.array import random_posdef_matrix
    >>> random_posdef_matrix(3, 0.1)
    ...
    """
    A = np.random.rand(N, N)
    return A @ A.T + alpha*np.eye(N)


@njit(nogil=True, parallel=True, cache=__cache)
def repeat(a: ndarray, N:int=1) -> ndarray:
    """
    Repeats an array N-times.
    
    Parameters
    ----------
    a : numpy.ndarray
        Input array.
    
    N : int, Optional
        Number of repetitions. Default is 1.
    
    Returns
    -------
    numpy.ndarray
        A NumPy array with shape (N, a.shape[0], a.shape[1]).
    
    Example
    -------
    
    For example, to generate basis vectors for 10 vectors embedded
    in the same coordinate frame, we need to stack up 10 identical
    identity matrices. This can be done the quickest by:
    
    >>> from neumann.array import repeat
    >>> axes = repeat(np.eye(3), 10)
    
    """
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat1d(a: ndarray, N=1) -> ndarray:
    """
    Repeats a 1d array N times.
    """
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i*M : (i+1)*M] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile(a: ndarray, da: ndarray, N=1) -> ndarray:
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a + i*da
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile1d(a: ndarray, da: ndarray, N=1) -> ndarray:
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i*M : (i+1)*M] = a + i*da
    return res


def indices_of_equal_rows(x: ndarray, y: ndarray, tol=1e-12):
    nX, dX = x.shape
    nY, dY = y.shape
    assert dX == dY, "Input arrays must have identical second dimensions."
    square = nX == nY
    integer = isintegerarray(x) and isintegerarray(y)
    if integer:
        if square:
            inds = np.flatnonzero((x == y).all(1))
            return inds, np.copy(inds)
        else:
            return indices_of_equal_rows_njit(x, y, 0)
    else:
        if square:
            return indices_of_equal_rows_square_njit(x, y, tol)
        else:
            return indices_of_equal_rows_njit(x, y, tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_njit(x: ndarray, y: ndarray, 
                               tol:float=1e-12):
    R = np.zeros((x.shape[0], y.shape[0]), dtype=x.dtype)
    for i in prange(R.shape[0]):
        for j in prange(R.shape[1]):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_square_njit(x: ndarray, y: ndarray,
                                      tol=1e-12):
    n = np.min([x.shape[0], y.shape[0]])
    R = np.zeros((n, n), dtype=x.dtype)
    for i in prange(n):
        for j in prange(n):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def count_cols(arr: ndarray) -> ndarray:
    """
    Count and return the number of columns for each row in
    the input array.
    """
    n = len(arr)
    res = np.zeros(n, dtype=np.int64)
    for i in prange(n):
        res[i] = len(arr[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat_diagonal_2d(a: ndarray, N:int=2) -> ndarray:
    """
    Assembles a 2d dense block-diagonal matrix by repeating.
    
    Parameters
    ----------
    a : numpy.ndarray
        A 2d array. This is the block that gets repeated.
    
    N : int, Optional
        The number of copies.
        
    Returns
    -------
    numpy.ndarray
        A 2d numpy array.
        
    """
    nR, nC = a.shape[:2]
    res = np.zeros((nR * N, nC * N), dtype=a.dtype)
    for i in prange(N):
        _iR = i * nR
        iR_ = _iR + nR
        _iC = i * nC
        iC_ = _iC + nC
        res[_iR:iR_, _iC:iC_] = a
    return res


# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find1d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = np.where(arr[i] == space)[0][0]
    return res


# !TODO generalize this up to n dimensions
# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find2d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = find1d(arr[i], space)
    return res


# -*- coding: utf-8 -*-
from sympy.parsing.sympy_parser import parse_expr
from sympy import lambdify, derive_by_array, symbols, Expr
from sympy.core.numbers import One
from collections import OrderedDict

from dewloosh.core.abc import ABCMeta_Weak


class ABCMeta_MetaFunction(ABCMeta_Weak):
    """
    Metaclass for defining ABCs for algebraic structures.
    """

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace,
                              *args, **kwargs)
        if 'value' in namespace:
            cls.f = namespace['value']
            
        if 'gradient' in namespace:
            cls.g = namespace['gradient']
            
        if 'Hessian' in namespace:
            cls.G = namespace['Hessian']
        
        return cls 


class MetaFunction(metaclass=ABCMeta_MetaFunction):
    __slots__ = ('f0', 'f1', 'f2', 'dimension', 'domain', 'expr',
                 'variables', 'vmap')

    def __call__(self, *args, **kwargs):
        return self.f0(*args, **kwargs)

    def f(self, *args, **kwargs):
        """
        Returns the function value.

        For this operation the object must have an implementation of

        value(self, *args, **kwargs):
            <...>
            return <...>
        """
        try:
            return self.f0(*args, **kwargs)
        except Exception:
            return None

    def g(self, *args, **kwargs):
        """
        Returns the gradient vector if available.

        For this operation the object must have an implementation of

        gradient(self, *args, **kwargs):
            <...>
            return <...>
        """
        try:
            return self.f1(*args, **kwargs)
        except Exception:
            return None

    def G(self, *args, **kwargs):
        """
        Returns the Hessian matrix if available.

        For this operation the object must have an implementation of

        Hessian(self,*args,**kwargs):
            <...>
            return <...>
        """

        try:
            return self.f2(*args, **kwargs)
        except Exception:
            return None

    @classmethod
    def _str_to_func(cls, str_expr: str, *args, **kwargs):
        return symbolize(*args, str_expr=str_expr, **kwargs)

    @classmethod
    def _sympy_to_func(cls, expr: Expr, *args, **kwargs):
        return symbolize(*args, expr=expr, **kwargs)


def decode(*args, expr=None, str_expr: str = None, variables=None,
           **kwargs):
    try:
        if str_expr is not None:
            expr = parse_expr(str_expr, evaluate=False)
        if not variables:
            variables = []
            """
            for arg in expr.args:
                syms = list(arg.free_symbols)
                if len(syms) == 1:
                    variables.append(syms[0])
            """
            variables = tuple(expr.free_symbols)
        else:
            try:
                variables = list(symbols(variables))
            except Exception:
                pass
        return expr, variables
    except Exception:
        return None, None


def symbolize(*args, **kwargs):
    expr, variables = decode(*args, **kwargs)
    f0 = lambdify([variables], expr, 'numpy')
    g = derive_by_array(expr, variables)
    f1 = lambdify([variables], g, 'numpy')
    G = derive_by_array(g, variables)
    f2 = lambdify([variables], G, 'numpy')
    return {'value': f0, 'gradient': f1,
            'Hessian': f2, 'd': len(variables),
            'variables': variables,
            'expr': expr}


def substitute(expr, values, variables=None, as_string=False):
    if variables is None:
        variables = tuple(expr.free_symbols)
    if not as_string:
        return expr.subs([(v, val) for v, val in zip(variables, values)])
    else:
        return expr.subs([(str(v), val) for v, val in zip(variables, values)])


def coefficients(expr=None, variables=None, normalize=False):
    try:
        if variables is None:
            variables = tuple(expr.free_symbols)
        d = OrderedDict({x: 0 for x in variables})
        d.update(expr.as_coefficients_dict())
        if not normalize:
            return d
        else:
            res = OrderedDict()
            for key, value in d.items():
                if len(key.free_symbols) == 0:
                    res[One()] = value*key
                else:
                    res[key] = value
            return res
    except Exception:
        return None


if __name__ == '__main__':

    str_expr = 'x*y + y**2 + 6*b + 2'
    d = decode(str_expr=str_expr)



""  # -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import norm
from collections.abc import Iterable

from ..function import Function


class MLSWeightFunction(Function):

    def __init__(self, core):

        if not isinstance(core, np.ndarray):
            if isinstance(core, Iterable):
                core = np.array(core)
                dim = core.shape[0]
            elif isinstance(core, int):
                dim = core
                core = np.zeros([dim])
            else:
                raise
        else:
            dim = core.shape[0]
        super().__init__(d=dim)
        self.core = core
        return

    def value(self, x):
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise
        return


def isMLSWeightFunction(f):
    c1 = isinstance(f, MLSWeightFunction)
    c2 = MLSWeightFunction in list(type(f).__bases__)
    return any([c1, c2])


class ConstantWeightFunction(MLSWeightFunction):

    def __init__(self, dimension, val=1.0):
        super().__init__(np.zeros([dimension]))
        self.val = val
        return

    def value(self, x):
        return self.val

    def gradient(self, x):
        return np.zeros([self.dimension])

    def Hessian(self, x):
        return np.zeros([self.dimension, self.dimension])


class SingularWeightFunction(MLSWeightFunction):

    def __init__(self, core, eps=1e-5):
        super().__init__(core)
        self.eps = eps
        return

    def value(self, x):
        super().value(x)
        return 1/(norm(np.subtract(self.core, x))**2 + self.eps**2)


class CubicWeightFunction(MLSWeightFunction):

    def __init__(self, core, supportdomain):
        super().__init__(core)
        assert self.dimension == 2
        self.supportdomain = supportdomain
        return

    def inner(self, x):

        d = np.subtract(self.core, x)
        difX = d[0]
        difY = d[1]
        dmX = self.supportdomain[0]
        dmY = self.supportdomain[1]
        rX = abs(difX)/dmX
        rY = abs(difY)/dmY

        if abs(difX) < 1e-12:
            drdX = 0
        else:
            drdX = (difX/abs(difX))/dmX

        if abs(difY) < 1e-12:
            drdY = 0
        else:
            drdY = (difY/abs(difY))/dmY

        if rX <= 0.5:
            wX = 2/3 - 4*rX**2 + 4*rX**3
            dwXdX = (-8*rX+12*rX**2)*drdX
            dwXdXX = (-8+24*rX)*drdX*drdX
        elif rX > 0.5 and rX <= 1:
            wX = 4/3 - 4*rX + 4*rX**2 - (4/3)*rX**3
            dwXdX = (-4+8*rX-4*rX**2)*drdX
            dwXdXX = (8-8*rX)*drdX*drdX
        else:
            wX = 0
            dwXdX = 0
            dwXdXX = 0

        if rY <= 0.5:
            wY = 2/3 - 4*rY**2 + 4*rY**3
            dwYdY = (-8*rY+12*rY**2)*drdY
            dwYdYY = (-8+24*rY)*drdY*drdY
        elif rY > 0.5 and rY <= 1:
            wY = 4/3 - 4*rY + 4*rY**2 - (4/3)*rY**3
            dwYdY = (-4+8*rY-4*rY**2)*drdY
            dwYdYY = (8-8*rY)*drdY*drdY
        else:
            wY = 0
            dwYdY = 0
            dwYdYY = 0

        val = wX*wY
        grad = np.array([wY*dwXdX, wX*dwYdY])
        Hessian = np.array([[wY*dwXdXX, dwXdX*dwYdY],
                            [dwXdX*dwYdY, wX*dwYdYY]])

        return val, grad, Hessian

    def value(self, x):
        super().value(x)
        res, _, _ = self.inner(x)
        return res

    def gradient(self, x):
        super().value(x)
        _, res, _ = self.inner(x)
        return res

    def Hessian(self, x):
        super().value(x)
        _, _, res = self.inner(x)
        return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numpy.linalg import LinAlgError
import sympy as sy
from sympy.utilities.iterables import multiset_permutations
from copy import copy, deepcopy
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from collections import defaultdict
from typing import Tuple
from enum import Enum, auto, unique
from time import time

from dewloosh.core.tools import getasany

from ..function import Function, InEquality, Equality, VariableManager
from ..function.relation import Relations, Relation
from ..function.metafunction import coefficients
from .errors import DegenerateProblemError, NoSolutionError
from ..array import atleast2d


__all__ = ['LinearProgrammingProblem']


@unique
class LinearProgrammingResult(Enum):
    UNIQUE = auto()
    MULTIPLE = auto()
    NOSOLUTION = auto()
    DEGENERATE = auto()
    

class LinearProgrammingProblem:
    """
    A lightweight class to handle general linear programming problems. It gaps the 
    bridge between the general form and the standard form. The class accepts
    symbolic expressions, but this should not be expected to be too fast. 
    For problems starting from medium size, it is suggested to use the 
    `solve_standard_form` method of the class.

    Parameters
    ----------
    constraints : Iterable[Function]
        List of constraint functions.

    variables : Iterable
        List of variables of the system.

    positive : bool, Optional
        A `True` value indicated that all variables are expected to take
        only positive values. Default is `True`.

    'obj', 'cost', 'payoff', 'fittness', 'f' : Function
        The objective function.

    Examples
    --------
    The examples requires `sympy` to be installed.

    (1) Example for unique solution.

    .. math::
        :nowrap:

        \\begin{eqnarray}
            & minimize&  \quad  3 x_1 + x_2 + 9 x_3 + x_4  \\\\
            & subject \, to& & \\\\
            & & x_1 + 2 x_3 + x_4 \,=\, 4, \\\\
            & & x_2 + x_3 - x_4 \,=\, 2, \\\\
            & & x_i \,\geq\, \, 0, \qquad i=1, \ldots, 4.
        \\end{eqnarray}

    >>> from neumann.optimize import LinearProgrammingProblem as LPP
    >>> import sympy as sy
    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> obj1 = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
    >>> eq11 = Equality(x1 + 2*x3 + x4 - 4, variables=syms)
    >>> eq12 = Equality(x2 + x3 - x4 - 2, variables=syms)
    >>> problem = LPP(cost=obj1, constraints=[eq11, eq12], variables=syms)
    >>> problem.solve()
    array([0., 6., 0., 4.])

    (2) Example for degenerate solution.

    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> obj2 = Function(3*x1 + x2 + 9*x3 + x4, variables=syms)
    >>> eq21 = Equality(x1 + 2*x3 + x4, variables=syms)
    >>> eq22 = Equality(x2 + x3 - x4 - 2, variables=syms)
    >>> P2 = LPP(cost=obj2, constraints=[eq21, eq22], variables=syms)
    >>> try:
    >>>     P2.solve()
    >>> except DegenerateProblemError:
    >>>     pass

    (3) Example for no solution.

    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> obj3 = Function(-3*x1 + x2 + 9*x3 + x4, variables=syms)
    >>> eq31 = Equality(x1 - 2*x3 - x4 + 2, variables=syms)
    >>> eq32 = Equality(x2 + x3 - x4 - 2, variables=syms)
    >>> P3 = LPP(cost=obj3, constraints=[eq31, eq32], variables=syms)
    >>> try:
    >>>     P3.solve()
    >>> except NoSolutionError:
    >>>     pass

    (4) Example for multiple solutions. We can ask for all the results
        with the option `return_all=True`.

    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> obj4 = Function(3*x1 + 2*x2 + 8*x3 + x4, variables=syms)
    >>> eq41 = Equality(x1 - 2*x3 - x4 + 2, variables=syms)
    >>> eq42 = Equality(x2 + x3 - x4 - 2, variables=syms)
    >>> P4 = LPP(cost=obj4, constraints=[eq41, eq42], variables=syms)
    >>> P4.solve(return_all=True)
    array([[0., 4., 0., 2.], [0., 1., 1., 0.]])

    """
    __tmpl_shift__ = 'y_{}'
    __tmpl_slack__ = 'z_{}'

    def __init__(self, *args, constraints=None, variables=None,
                 positive=None, standardform=False, **kwargs):
        super().__init__()
        self.obj = None
        self.constraints = None
        self.standardform = None
        self.vmanager = None
        self._hook = None
        self._update(*args,
                     constraints=constraints,
                     variables=variables,
                     positive=positive,
                     standardform=standardform, **kwargs)

    def _update(self, *args, variables=None, positive=None, **kwargs):
        self.standardform = kwargs.get('standardform', self.standardform)
        if len(args) > 0:
            obj = None
            if isinstance(args[0], Function):
                obj = args[0]
            if obj is not None:
                self.obj = obj
        if self.obj is None:
            self.obj = getasany(['obj', 'cost', 'payoff', 'fittness', 'f'],
                                None, **kwargs)
        assert self.obj is not None, "Objective must be set on instance creation!"
        self.constraints = kwargs.get('constraints', self.constraints)
        if self.constraints is None:
            self.constraints = []
        if variables is not None:
            if isinstance(positive, bool):
                self.vmanager = VariableManager(variables, positive=positive)
            else:
                self.vmanager = VariableManager(variables)
        self._hook = kwargs.get('_hook', self._hook)

    @property
    def vm(self):
        return self.vmanager

    @staticmethod
    def example_unique() -> 'LinearProgrammingProblem':
        """
        Returns the following LPP:

        .. math::
            :nowrap:

            \\begin{eqnarray}
                & minimize&  \quad  3 x_1 + x_2 + 9 x_3 + x_4  \\\\
                & subject \, to& & \\\\
                & & x_1 + 2 x_3 + x_4 \,=\, 4, \\\\
                & & x_2 + x_3 - x_4 \,=\, 2, \\\\
                & & x_i \,\geq\, \, 0, \qquad i=1, \ldots, 4.
            \\end{eqnarray}

        The returned LPP has a unique solution:

        :math:`\quad \mathbf{x} = (0., 6., 0., 4.), \quad f(\mathbf{x}) = 10.`

        Example
        -------
        >>> from neumann.optimize import LinearProgrammingProblem as LPP
        >>> problem = LPP.example_unique()
        >>> problem.solve()['x']
        array([0., 6., 0., 4.])

        """
        variables = ['x1', 'x2', 'x3', 'x4']
        x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
        obj1 = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
        eq11 = Equality(x1 + 2*x3 + x4 - 4, variables=syms)
        eq12 = Equality(x2 + x3 - x4 - 2, variables=syms)
        P = LinearProgrammingProblem(cost=obj1, constraints=[eq11, eq12],
                                     variables=syms, positive=True)
        return P

    def add_constraint(self, *args, **kwargs):
        """
        Adds a new constraint to the system.
        """
        if isinstance(args[0], Function):
            if isinstance(args[0], InEquality):
                assert args[0].op in [Relations.ge, Relations.le], \
                    "Only '>=' and '<=' arre allowed!"
            c = args[0]
        else:
            c = Relation(*args, **kwargs)
        self.constraints.append(c)

    @property
    def variables(self):
        return self.vmanager.target()

    def _sync_variables(self):
        s = set()
        s.update(self.obj.variables)
        for c in self.constraints:
            s.update(c.variables)
        self.vmanager.add_variables(s)

    def _shift_variables(self):
        """
        Handle variables not restricted in sign.
        """
        vmap = dict()
        tmpl = self.__class__.__tmpl_shift__
        count = 1
        for v in self.vmanager.source():
            if not v.is_positive:
                sym = [tmpl.format(count), tmpl.format(count + 1)]
                si, sj = sy.symbols(sym, positive=True)
                vmap[v] = si - sj
        self.vmanager.substitute(vmap)

    def get_system_variables(self) -> list:
        """
        Returns all variables of the system.
        """
        s = set()
        s.update(self.obj.variables)
        for c in self.constraints:
            s.update(c.variables)
        return list(s)

    def get_slack_variables(self, template=None) -> list:
        """
        Returns the slack variables of the inequality constraints of the problem.
        """
        tmpl = self.__class__.__tmpl_slack__ if template is None else template
        c = self.constraints
        inequalities = list(filter(lambda c: isinstance(c, InEquality), c))
        n = len(inequalities)
        strlist = list(map(tmpl.format, range(1, n + 1)))
        y = sy.symbols(strlist, positive=True)
        for i in range(n):
            inequalities[i].slack = y[i]
        return y

    def has_standardform(self) -> bool:
        """
        Tells if the object admits the standard form of a linear program or not.

        Notes
        -----
        An LPP admits standard form if
            * only equality constraints are present
            * all variables are positive

        Returns
        -------
        bool
            `True` if the problem admits the standard form.

        """
        all_eq = all([isinstance(c, Equality) for c in self.constraints])
        all_pos = all([v.is_positive for v in self.get_system_variables()])
        return all_pos and all_eq

    def simplify(self, maximize=False, inplace=False) -> 'LinearProgrammingProblem':
        """
        Simplifies the problem so that it admits the standard form. 

        This is done in 3 steps:

        1) Decomposing variables not restricted in sign according to
        
            :math:`x = x^{+} - x^{-}, \quad \mid x \mid = x^{+} + x^{-}`, 
            
            where 
            
            :math:`x^{+} = max(0, x) \geq 0, \quad x^{-} = max(0, -x) \geq 0`.
            
            Then the variable :math:`x` can be substituted by         
            
            :math:`x = x^{+} - x^{-}` and the conditions :math:`\quad x^{+}, x^{-} \geq 0`.
            
        2) Transforming inequalities into equalities using slack variables.
            
            We introduce new variables
            
            :math:`\mathbf{y} = \mathbf{b} - \mathbf{A} \mathbf{x} \geq \mathbf{0}`,
            
            and formulate the extended problem
            
            :math:`\hat{\mathbf{A}} \mathbf{X} = \mathbf{b}`,
            
            where 
            
            :math:`\mathbf{X} = (\mathbf{x} \, \, \mathbf{y}), \quad \hat{\mathbf{A}} = (\mathbf{A} \, \, \mathbf{1})`.
            
        3) Transform to minimization problem if necessary using the simple rule
            
            :math:`max(f) = -min(-f)`.

        Parameters
        ----------
        maximize : bool
            Set this true if the problem is a maximization. Default is False.

        inplace : bool
            If `True`, transformation happend in place, changing the internal structure
            ob the instance it is invoked by. In this case, `self` gets returned for
            possible continuation.

        Notes
        -----
        Steps 1) and 2) both increase the number of variables of the standard system.

        Returns
        -------
        LinearProgrammingProblem
            An LPP that admits a standard form.

        dict, Optional
            A mapping between variables of the standard and the general form.
            Only if `return_inverse` is `True`.

        Raises
        ------
        NotImplementedError
            On invalid input.

        """
        P = self if inplace else deepcopy(self)
        # handle variables
        P._sync_variables()
        P._shift_variables()
        s = P.get_slack_variables()
        P.vmanager.add_variables(s)
        v = P.vmanager.source()  # can be in arbitrary order
        n = len(v)
        x = list(sy.symbols(' '.join(['X_{}'.format(i) for i in range(1, n + 1)]),
                            positive=True))
        c = list(sy.symbols(
            ' '.join(['c_{}'.format(i) for i in range(1, n + 1)])))
        x_v = {x_: v_ for x_, v_ in zip(x, v)}
        P.vmanager.substitute(vmap=x_v, inverse=True)
        v.append(1)
        x.append(1)
        c.append(sy.symbols('c'))
        x_c = {x_: c_ for x_, c_ in zip(x, c)}
        template = np.inner(c, x)
        x.pop(-1)
        v.pop(-1)
        vmap = P.vmanager.vmap  # == v_x, inverse of x_v
        smap = {x_v[vmap[s]]: 1 for s in s}

        def redefine_expr(fnc, aux: dict = None):
            expr = fnc.expr.subs([(v, expr) for v, expr in vmap.items()])
            fnc_coeffs = coefficients(expr=expr, normalize=True)
            coeffs = defaultdict(lambda: 0)
            if isinstance(aux, dict):
                coeffs.update(aux)
            coeffs.update({x_c[x]: c for x, c in fnc_coeffs.items()})
            return template.subs([(c_, coeffs[c_]) for c_ in c])

        def redefine_function(fnc):
            minmax = -1 if maximize else 1
            expr = minmax * redefine_expr(fnc)
            return Function(expr, variables=x, vmap=x_v)

        def redefine_equality(fnc):
            expr = redefine_expr(fnc)
            return Equality(expr, variables=x, vmap=x_v)

        def redefine_inequality(fnc):
            expr = redefine_expr(fnc, smap)
            if fnc.op == Relations.ge:
                pass
            elif fnc.op == Relations.le:
                expr *= -1
            else:
                raise NotImplementedError("Only >= and <= are allowed!")
            expr -= vmap[fnc.slack]
            eq = Equality(expr, variables=x, vmap=x_v)
            eq.slack = fnc.slack
            return eq

        obj = redefine_function(P.obj)
        _c = P.constraints
        constraints = []
        eq = list(map(redefine_equality, filter(
            lambda c: isinstance(c, Equality), _c)))
        constraints += eq
        if len(smap) > 0:
            ieq = list(map(redefine_inequality, filter(
                lambda c: isinstance(c, InEquality), _c)))
            constraints += ieq
        if inplace:
            self._update(constraints=constraints,
                         variables=x, positive=True, _hook=P)
        else:
            lpp = LinearProgrammingProblem(obj=obj, constraints=constraints,
                                           variables=x, positive=True, _hook=P)
            lpp.vmanager.inv = x_v
            return lpp

    def eval_constraints(self, x: Iterable) -> Iterable:
        """
        Evaluates the constraints at `x`.
        """
        return np.array([c.f0(x) for c in self.constraints], dtype=float)

    def feasible(self, x: Iterable = None) -> bool:
        """
        Returns `True` if `x` is a feasible candidate to the current problem,
        `False` othwerise.
        """
        c = [c.relate(x) for c in self.constraints]
        if self.has_standardform():
            return all(c) and all(x >= 0)
        else:
            return all(c)

    @staticmethod
    def basic_solution(A=None, b=None, order=None) -> Tuple[ndarray]:
        """
        Returns a basic solution to a problem the form

        .. math::
            :nowrap:

            \\begin{eqnarray}
                minimize  \quad  \mathbf{c}\mathbf{x} \quad under \quad 
                \mathbf{A}\mathbf{x}=\mathbf{b}, \quad \mathbf{x} \, \geq \, 
                \mathbf{0}.
            \\end{eqnarray}

        where :math:`\mathbf{b} \in \mathbf{R}^m, \mathbf{c} \in \mathbf{R}^n` and :math:`\mathbf{A}` is
        an :math:`m \\times n` matrix with :math:`n>m`.

        Parameters
        ----------
        A : numpy.ndarray
            An :math:`m \times n` matrix with :math:`n>m`

        b : numpy.ndarray
            Right-hand sides. :math:`\mathbf{b} \in \mathbf{R}^m`

        order : Iterable[int], Optional
            An arbitrary permutation of the indices.

        Returns
        -------
        numpy.ndarray 
            Coefficient matrix :math:`\mathbf{B}`

        numpy.ndarray
            Inverse of coefficient matrix :math:`\mathbf{B}^{-1}`

        numpy.ndarray
            Coefficient matrix :math:`\mathbf{N}`

        numpy.ndarray
            Basic solution :math:`\mathbf{x}_{B}`

        numpy.ndarray
            Remaining solution :math:`\mathbf{x}_{N}`

        """
        m, n = A.shape
        r = n - m
        assert r > 0

        stop = False
        try:
            if order is not None:
                if isinstance(order, Iterable):
                    permutations = iter([order])
            else:
                order = [i for i in range(n)]
                permutations = multiset_permutations(order)
            while not stop:
                order = next(permutations)
                A_ = A[:, order]
                B_ = A_[:, :m]
                try:
                    B_inv = np.linalg.inv(B_)
                    xB = np.matmul(B_inv, b)
                    stop = all(xB >= 0)
                except LinAlgError:
                    """
                    If there is no error, it means that calculation
                    of xB was succesful, which is only possible if the
                    current permutation defines a positive definite submatrix.
                    Note that this is cheaper than checking the eigenvalues,
                    since it only requires the eigenvalues to be all positive,
                    and does not involve calculating their actual values.
                    """
                    pass
        except StopIteration:
            """
            There is no permutation of columns that would produce a regular
            mxm submatrix
                -> there is no feasible basic solution
                    -> there is no feasible solution
            """
            pass
        finally:
            if stop:
                N_ = A_[:, m:]
                xN = np.zeros(r, dtype=float)
                return B_, B_inv, N_, xB, xN, order
            else:
                return None

    @staticmethod
    def solve_standard_form(A: ndarray, b: ndarray, c: ndarray, order=None,
                            tol: float = 1e-10):
        """Solves a linear problem in standard form:

        :math:`minimize \quad \mathbf{c} \mathbf{x} \quad under \quad \mathbf{A}\mathbf{x} = \mathbf{b}`.

        See the notes section for the behaviour and the possible gotchas.

        Parameters
        ----------
        A : numpy.ndarray
            2d float array.

        b : numpy.ndarray
            1d float array.

        c : numpy.ndarray
            1d float array.

        order : Iterable, Optional
            The order of the variables.

        tol : float, Optional
            Floating point tolerance. Default is 1e-10.

        Returns
        -------
        numpy.ndarray
            Results as a 1d (unique solution) or a 2d (multiple solutions) 
            numpy array.

        Notes
        -----
        1) The value of the parameter `tol` is used to make judgements on the vanishing ratios 
        of entering variables, therfore effects the detection of degenerate situations. The higher
        the value, the more tolerant the system is to violations.

        2) The line between the unique, the degenerate situation and having no solution at all may 
        be very thin in some settings. In such a scenario, repeated solutions might return a 
        a solution, a `NoSolutionError` or a `DegenerateProblemError` as well. Problems
        with this behaviour are all considered degenerate, and suggest an ill-posed setup.

        Raises
        ------
        NoSolutionError
            If there is no solution to the problem.

        DegenerateProblemError
            If the problem is degenerate.

        """
        m, n = A.shape
        r = n - m
        assert r > 0
        basic = LinearProgrammingProblem.basic_solution(A, b, order=order)
        if basic:
            B, B_inv, N, xB, xN, order = basic
            c_ = c[order]
            cB = c_[:m]
            cN = c_[m:]
            t = None
        else:
            raise NoSolutionError("Failed to find basic solution!")

        def unit_basis_vector(length, index=0, value=1.0):
            return value * np.bincount([index], None, length)

        def enter(i_enter):
            nonlocal B, B_inv, N, xB, xN, order, cB, cN, t
            b_enter = unit_basis_vector(r, i_enter, 1.0)

            # w = vector of decrements of the current solution xB
            # Only positive values are a threat to feasibility, and we
            # need to tell which of the components of xB vanishes first,
            # which, since all components of xB are posotive,
            # has to do with the positive components only.
            w_enter = np.matmul(W, b_enter)
            i_leaving = np.argwhere(w_enter > 0)
            if len(i_leaving) == 0:
                # step size could be indefinitely increased in this
                # direction without violating feasibility, there is
                # no solution to the problem
                raise NoSolutionError("There is not solution to this problem!")

            vanishing_ratios = xB[i_leaving]/w_enter[i_leaving]
            # the variable that vanishes first is the one with the smallest
            # vanishing ratio
            i_leave = i_leaving.flatten()[np.argmin(vanishing_ratios)]

            # step size in the direction of current basis vector
            t = xB[i_leave]/w_enter[i_leave]

            # update solution
            if abs(t) <= tol:
                # Smallest vanishing ratio is zero, any step would
                # result in an infeasible situation.
                # -> go for the next entering variable
                return False
            xB -= t*w_enter
            xN = t*b_enter

            order[m + i_enter], order[i_leave] = \
                order[i_leave], order[m + i_enter]
            B[:, i_leave], N[:, i_enter] = \
                N[:, i_enter], copy(B[:, i_leave])
            B_inv = np.linalg.inv(B)
            cB[i_leave], cN[i_enter] = cN[i_enter], cB[i_leave]
            xB[i_leave], xN[i_enter] = xN[i_enter], xB[i_leave]
            return True

        def unique_result():
            return np.concatenate((xB, xN))[np.argsort(order)]

        def multiple_results():
            assert np.all(reduced_costs >= 0)
            assert reduced_costs.min() <= tol
            inds = np.where(reduced_costs <= tol)[0]
            res = [unique_result(), ]
            for i in inds:
                assert enter(i)
                res.append(unique_result())
            return np.stack(res)

        degenerate = False
        while True:
            if degenerate:
                # The objective could be decreased, but only on the expense
                # of violating positivity of the standard variables.
                # Hence, the solution is degenerate.
                raise DegenerateProblemError('The problem is ill posed!')

            # calculate reduced costs
            W = np.matmul(B_inv, N)
            reduced_costs = cN - np.matmul(cB, W)
            nEntering = np.count_nonzero(reduced_costs < 0)
            if nEntering == 0:
                # The objective can not be further reduced.
                # There was only one basic solution, which is
                # a unique optimizer.
                d = np.count_nonzero(reduced_costs >= tol)
                if d < len(reduced_costs):
                    # there are edges along with the objective does
                    # not increase
                    #dc = np.abs(reduced_costs - reduced_costs.min())
                    #inds = np.where(dc <= tol)[0]
                    return multiple_results()
                else:
                    return unique_result()
            # If we reach this line, reduction of the objective is possible,
            # although maybe indefinitely. If the objective can be decreased,
            # but only on the expense of violating feasibility, the
            # solution is degenerate.
            degenerate = True

            # Candidates for entering index are the indices of the negative
            # components of the vector of reduced costs.
            i_entering = np.argsort(reduced_costs)[:nEntering]
            for i_enter in i_entering:
                if not enter(i_enter):
                    # Smallest vanishing ratio is zero, any step would
                    # result in an infeasible situation.
                    # -> go for the next entering variable
                    continue

                # break loop at the first meaningful (t != 0) decrease and
                # force recalculation of the vector of reduced costs
                degenerate = False
                break

    def to_numpy(self, maximize:bool=False, return_coeffs:bool=False):
        """
        Returns the complete numerical representation of the standard 
        form of the problem:

            :math:`minimize \quad \mathbf{c} \mathbf{x} \quad under \quad \mathbf{A}\mathbf{x} = \mathbf{b}`.

        Parameters
        ----------
        maximize : bool
            Set this true if the problem is a maximization. Default is False.

        return_coeffs : bool
            If `True`, linear coefficients of the design variables are returned as
            a sequence of `SymPy` symbols.

        Returns
        -------
        numpy.ndarray
            2d NumPy array 'A'

        numpy.ndarray
            1d NumPy array 'b'

        list, Optional
            A list of `SymPy` symbols. Only if `return_coeffs` is `True`.

        """
        if not self.has_standardform():
            P = self.simplify(maximize=maximize)
        else:
            P = self
        x = P.variables
        n = len(x)
        zeros = np.zeros((n,), dtype=float)
        b = - P.eval_constraints(zeros)
        A = []
        for c in P.constraints:
            coeffs = c.linear_coefficients(normalize=True)
            A.append(np.array([coeffs[x_] for x_ in x], dtype=float))
        A = np.vstack(A)
        coeffs = P.obj.linear_coefficients(normalize=True)
        if return_coeffs:
            coeffs = P.obj.linear_coefficients(normalize=True)
            c = np.array([coeffs[x_] for x_ in x], dtype=float)
            return A, b, c
        return A, b

    def maximize(self, *args, **kwargs):
        """
        Solves the LPP as a maximization. For the possible arguments, see `solve`.

        """
        kwargs['maximize'] = True
        return self.solve(*args, **kwargs)

    def solve(self, order=None, return_all:bool=True, maximize:bool=False,
              as_dict:bool=False, raise_errors:bool=False, tol: float = 1e-10):
        """
        Solves the problem and return the solution(s) if there are any.

        Parameters
        ----------
        order : Iterable, Optional
            The order of the variables. This might speed up finding the
            basic solution. Default is None.

        as_dict: bool
            If `True`, the results are returned as a dictionary, where the
            keys are sympy symbols of the variables of the problem.
        
        raise_errors : bool
            If `True`, the solution raises the actual errors on exception events,
            otherwise they get returned within the result, under key `e`.
        
        tol : float, Optional
            Floating point tolerance. Default is 1e-10.

        Notes
        -----
        It is assumed that this function gets invoked on relatively small problems.
        For large-scale situations, we suggest to use `solve_standard_form` function
        of this class instead. Contrary to this one, it rases errors, while the more
        high-level `solve` method of the instance catches those errors, and returns
        them as part of the returned dictionary.

        Returns
        -------
        dict
            A dictionary with the following items:
            
            x : numpy.ndarray or dict
                The solution as an array or a dictionary, depending on your input.
            
            e : Iterable
                A list of errors that occured during solution.
            
            time : dict
                A dictionary with information of execution times of the main stages
                of the calculation.

        """
        summary = {'time' : {}, 'x' : None, 'e': []}
        try:
            # general form -> standard form
            _t0 = time()
            P = self.simplify(maximize=maximize, inplace=False)
            A, b, c = P.to_numpy(maximize=False, return_coeffs=True)
            summary['time']['preproc'] = time() - _t0

            # calculate solution
            _cls = self.__class__
            _t0 = time()
            x = _cls.solve_standard_form(A, b, c, order=order, tol=tol)
            summary['time']['solution'] = time() - _t0
            
            # standard form -> general form
            _t0 = time()
            res = None
            if x is not None:
                svars = P.variables  # standard variables
                gvars = self.variables  # general variables
                if not return_all:
                    x = x[0]
                x = atleast2d(x)
                if as_dict:
                    arr = []
                    gdata = {g: [] for g in gvars}
                    for i in range(x.shape[0]):
                        smap = {s: sx for s, sx in zip(svars, x[i])}
                        vm = P._hook.vm.substitute(smap, inplace=False)
                        [gdata[g].append(vm[g]) for g in gvars]
                    res = {g: np.squeeze(np.array(gdata[g], dtype=float)) for g in gvars}                
                else:
                    arr = []
                    for i in range(x.shape[0]):
                        smap = {s: sx for s, sx in zip(svars, x[i])}
                        vm = P._hook.vm.substitute(smap, inplace=False)
                        vals = [vm[g] for g in gvars]
                        arr.append(np.array(list(vals), dtype=float))
                    res = np.squeeze(np.stack(arr))                
            summary['x'] = res
            summary['time']['postproc'] = time() - _t0
        except Exception as e:
            summary['e'].append(e)
        finally:
            if len(summary['e']) > 0:
                summary['time']['solution'] = None
                if raise_errors:
                    raise summary['e'][0]    
            return summary
        


# -*- coding: utf-8 -*-
import codecs
import os.path
from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")
    
    
def get_description(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__description__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find description string.")


with open("README.md", "r") as fh:
    long_description = fh.read()


with open('requirements.txt') as f:
    required = f.read().splitlines()


_init_path = "src/neumann/__init__.py"
_version = get_version(_init_path)
_description = get_description(_init_path)
_url = 'https://github.com/dewloosh/Neumann'
_download_url = _url + '/archive/refs/tags/{}.zip'.format(_version)


setup(
	name="neumann",
    version=_version,                        
    author="Bence Balogh",
    author_email = 'dewloosh@gmail.com',                   
    description=_description,
    long_description=long_description,   
    long_description_content_type="text/markdown",
	url = _url, 
    download_url = _download_url,
	packages=find_packages(where='src'),
	classifiers=[
        'Development Status :: 5 - Production/Stable',     
        'License :: OSI Approved :: MIT License',   
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",   
        'Programming Language :: Python :: 3 :: Only',
		'Operating System :: OS Independent',
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics"
    ],                                      
    python_requires='>=3.7, <3.11',                             
    package_dir={'':'src'},     
    install_requires=required,
	zip_safe=False,
)



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange, guvectorize as guv
from numpy import ndarray
import sympy as sy
from sympy import symbols, Matrix

__cache = True


def inv_sym_3x3(m: Matrix, as_adj_det=False):
    P11, P12, P13, P21, P22, P23, P31, P32, P33 = \
        symbols('P_{11} P_{12} P_{13} P_{21} P_{22} P_{23} P_{31} \
                P_{32} P_{33}', real=True)
    Pij = [[P11, P12, P13], [P21, P22, P23], [P31, P32, P33]]
    P = sy.Matrix(Pij)
    detP = P.det()
    adjP = P.adjugate()
    invP = adjP / detP
    subs = {s: r for s, r in zip(sy.flatten(P), sy.flatten(m))}
    if as_adj_det:
        return detP.subs(subs), adjP.subs(subs)
    else:
        return invP.subs(subs)


@njit(nogil=True, parallel=True, cache=__cache)
def vpath(p1: ndarray, p2: ndarray, n: int):
    nD = len(p1)
    dist = p2 - p1
    length = np.linalg.norm(dist)
    s = np.linspace(0, length, n)
    res = np.zeros((n, nD), dtype=p1.dtype)
    d = dist / length
    for i in prange(n):
        res[i] = p1 + s[i] * d
    return res


@njit(nogil=True, cache=__cache)
def solve(A, b):
    return np.linalg.solve(A, b)


@njit(nogil=True, cache=__cache)
def inv(A: ndarray):
    return np.linalg.inv(A)


@njit(nogil=True, cache=__cache)
def matmul(A: ndarray, B: ndarray):
    return A @ B


@njit(nogil=True, cache=__cache)
def ATB(A: ndarray, B: ndarray):
    return A.T @ B


@njit(nogil=True, cache=__cache)
def matmulw(A: ndarray, B: ndarray, w: float = 1.0):
    return w * (A @ B)


@njit(nogil=True, cache=__cache)
def ATBA(A: ndarray, B: ndarray):
    return A.T @ B @ A


@njit(nogil=True, cache=__cache)
def ATBAw(A: ndarray, B: ndarray, w: float = 1.0):
    return w * (A.T @ B @ A)


@guv(['(f8[:, :], f8)'], '(n, n) -> ()', nopython=True, cache=__cache)
def det3x3(A, res):
    res = A[0, 0] * A[1, 1] * A[2, 2] - A[0, 0] * A[1, 2] * A[2, 1] \
        - A[0, 1] * A[1, 0] * A[2, 2] + A[0, 1] * A[1, 2] * A[2, 0] \
        + A[0, 2] * A[1, 0] * A[2, 1] - A[0, 2] * A[1, 1] * A[2, 0]


@guv(['(f8[:, :], f8)'], '(n, n) -> ()', nopython=True, cache=__cache)
def det2x2(A, res):
    res = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@njit(nogil=True, cache=__cache)
def inv2x2(A):
    res = np.zeros_like(A)
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = - A[0, 1] / d
    res[1, 0] = - A[1, 0] / d
    return res


@guv(['(f8[:, :], f8[:, :])'], '(n, n) -> (n, n)', nopython=True, cache=__cache)
def inv2x2u(A, res):
    d = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res[0, 0] = A[1, 1] / d
    res[1, 1] = A[0, 0] / d
    res[0, 1] = - A[0, 1] / d
    res[1, 0] = - A[1, 0] / d


@guv(['(f8[:, :], f8[:, :])'], '(n, n) -> (n, n)', nopython=True, cache=__cache)
def adj3x3(A, res):
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]


@guv(['(f8[:, :], f8[:, :])'], '(n, n) -> (n, n)', nopython=True, cache=__cache)
def inv3x3u(A, res):
    det = A[0, 0] * A[1, 1] * A[2, 2] - A[0, 0] * A[1, 2] * A[2, 1] \
        - A[0, 1] * A[1, 0] * A[2, 2] + A[0, 1] * A[1, 2] * A[2, 0] \
        + A[0, 2] * A[1, 0] * A[2, 1] - A[0, 2] * A[1, 1] * A[2, 0]
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res /= det


@njit(nogil=True, cache=__cache)
def inv3x3(A):
    res = np.zeros_like(A)
    det = A[0, 0] * A[1, 1] * A[2, 2] - A[0, 0] * A[1, 2] * A[2, 1] \
        - A[0, 1] * A[1, 0] * A[2, 2] + A[0, 1] * A[1, 2] * A[2, 0] \
        + A[0, 2] * A[1, 0] * A[2, 1] - A[0, 2] * A[1, 1] * A[2, 0]
    res[0, 0] = A[1, 1] * A[2, 2] - A[1, 2] * A[2, 1]
    res[0, 1] = -A[0, 1] * A[2, 2] + A[0, 2] * A[2, 1]
    res[0, 2] = A[0, 1] * A[1, 2] - A[0, 2] * A[1, 1]
    res[1, 0] = -A[1, 0] * A[2, 2] + A[1, 2] * A[2, 0]
    res[1, 1] = A[0, 0] * A[2, 2] - A[0, 2] * A[2, 0]
    res[1, 2] = -A[0, 0] * A[1, 2] + A[0, 2] * A[1, 0]
    res[2, 0] = A[1, 0] * A[2, 1] - A[1, 1] * A[2, 0]
    res[2, 1] = -A[0, 0] * A[2, 1] + A[0, 1] * A[2, 0]
    res[2, 2] = A[0, 0] * A[1, 1] - A[0, 1] * A[1, 0]
    res /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk(A):
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        det = A[i, 0, 0] * A[i, 1, 1] * A[i, 2, 2] - A[i, 0, 0] * A[i, 1, 2] \
            * A[i, 2, 1] - A[i, 0, 1] * A[i, 1, 0] * A[i, 2, 2] \
            + A[i, 0, 1] * A[i, 1, 2] * A[i, 2, 0] + A[i, 0, 2] \
            * A[i, 1, 0] * A[i, 2, 1] - A[i, 0, 2] * A[i, 1, 1] * A[i, 2, 0]
        res[i, 0, 0] = A[i, 1, 1] * A[i, 2, 2] - A[i, 1, 2] * A[i, 2, 1]
        res[i, 0, 1] = -A[i, 0, 1] * A[i, 2, 2] + A[i, 0, 2] * A[i, 2, 1]
        res[i, 0, 2] = A[i, 0, 1] * A[i, 1, 2] - A[i, 0, 2] * A[i, 1, 1]
        res[i, 1, 0] = -A[i, 1, 0] * A[i, 2, 2] + A[i, 1, 2] * A[i, 2, 0]
        res[i, 1, 1] = A[i, 0, 0] * A[i, 2, 2] - A[i, 0, 2] * A[i, 2, 0]
        res[i, 1, 2] = -A[i, 0, 0] * A[i, 1, 2] + A[i, 0, 2] * A[i, 1, 0]
        res[i, 2, 0] = A[i, 1, 0] * A[i, 2, 1] - A[i, 1, 1] * A[i, 2, 0]
        res[i, 2, 1] = -A[i, 0, 0] * A[i, 2, 1] + A[i, 0, 1] * A[i, 2, 0]
        res[i, 2, 2] = A[i, 0, 0] * A[i, 1, 1] - A[i, 0, 1] * A[i, 1, 0]
        res[i] /= det
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inv3x3_bulk2(A):
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = inv3x3(A[i])
    return res


@njit(nogil=True, cache=__cache)
def normalize(A):
    return A/np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def normalize2d(A):
    res = np.zeros_like(A)
    for i in prange(A.shape[0]):
        res[i] = normalize(A[i])
    return res


@njit(nogil=True, cache=__cache)
def norm(A):
    return np.linalg.norm(A)


@njit(nogil=True, parallel=True, cache=__cache)
def norm2d(A):
    res = np.zeros(A.shape[0])
    for i in prange(A.shape[0]):
        res[i] = norm(A[i, :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _to_range_1d(vals: ndarray, source: ndarray, target: ndarray):
    res = np.zeros_like(vals)
    s0, s1 = source
    t0, t1 = target
    b = (t1 - t0) / (s1 - s0)
    a = (t0 + t1) / 2 - b * (s0 + s1) / 2
    for i in prange(res.shape[0]):
        res[i] = a + b * vals[i]
    return res


def to_range_1d(vals: ndarray, *args, source: ndarray, target: ndarray = None,
             squeeze=False, **kwargs):
    if not isinstance(vals, ndarray):
        vals = np.array([vals, ])
    source = np.array([0., 1.]) if source is None else np.array(source)
    target = np.array([-1., 1.]) if target is None else np.array(target)
    if squeeze:
        return np.squeeze(_to_range_1d(vals, source, target))
    else:
        return _to_range_1d(vals, source, target)


@njit(nogil=True, parallel=True, cache=__cache)
def _linspace(p0: ndarray, p1: ndarray, N):
    s = p1 - p0
    L = np.linalg.norm(s)
    n = s / L
    djac = L/(N-1)
    step = n * djac
    res = np.zeros((N, p0.shape[0]))
    res[0] = p0
    for i in prange(1, N-1):
        res[i] = p0 + i*step
    res[-1] = p1
    return res


def linspace(start, stop, N):
    if isinstance(start, ndarray):
        return _linspace(start, stop, N)
    else:
        return np.linspace(start, stop, N)


@njit(nogil=True, parallel=True, cache=__cache)
def linspace1d(start, stop, N):
    res = np.zeros(N)
    di = (stop - start) / (N - 1)
    for i in prange(N):
        res[i] = start + i * di
    return res


# -*- coding: utf-8 -*-
import numpy as np
from numba import jit

from .frutils import max_frontwidth_bulk, flatsize_sym, flatind_sym
from .topo import signed_topo_bulk


@jit(nopython=True, nogil=True, cache=True)
def frontal_sym_bulk_uniform(A: np.ndarray, topology: np.ndarray,
                             B: np.ndarray, presc_bool: np.ndarray,
                             presc_val: np.ndarray, epath: np.ndarray):
    nEQ, nRHS = B.shape
    nE, nXE = topology.shape
    topology += 1
    topo_sig = signed_topo_bulk(topology, epath)
    maxwidth = max_frontwidth_bulk(topo_sig, epath)
    A_front = np.zeros(flatsize_sym(maxwidth), dtype=A.dtype)
    B_front = np.zeros((maxwidth, nRHS), dtype=A.dtype)
    front_to_glob = np.zeros(maxwidth, dtype=topology.dtype)
    loc_to_front = np.zeros(nXE, dtype=topology.dtype)
    lhs = np.zeros((nEQ, maxwidth), dtype=A.dtype)
    rhs = np.zeros((nEQ, nRHS), dtype=A.dtype)
    glob_to_width = np.zeros(nEQ, dtype=topology.dtype)
    glob_to_front = np.zeros(nEQ, dtype=topology.dtype)
    eqpath = np.zeros(nEQ, dtype=topology.dtype)
    frontwidth = 0

    # ASSEMBLY & ELIMINATION
    iEQ = 0
    for iE in epath:
        # find front indices for local variables
        for iXE in range(nXE):
            gind = topology[iE, iXE]
            frontindex = -1
            for ifront in range(frontwidth):
                if front_to_glob[ifront] == gind:
                    frontindex = ifront
                    break
            if frontindex < 0:
                for ifront in range(frontwidth):
                    if front_to_glob[ifront] == 0:
                        frontindex = ifront
                        break
            if frontindex < 0:
                frontindex = frontwidth
                frontwidth += 1
            front_to_glob[frontindex] = gind
            loc_to_front[iXE] = frontindex

        # local assembly
        nElim = 0
        for iXE in range(nXE):
            ifront = loc_to_front[iXE]
            for jXE in range(iXE, nXE):
                jfront = loc_to_front[jXE]
                index = flatind_sym(ifront, jfront)
                A_front[index] += A[iE, iXE, jXE]
            gind = - topo_sig[iE, iXE]
            if gind > 0:
                # because RHS is assumed to be present as a full array,
                # assembly happends only the last time the variable is
                # accessed
                B_front[ifront] += B[gind - 1]
                nElim += 1

        # local elimination
        cElim = 0
        for iXE in range(nXE):
            # check if variable needs to be handled
            if cElim == nElim:
                break
            ifront = loc_to_front[iXE]
            gind = - topo_sig[iE, iXE]
            if gind < 0:
                continue
            else:
                gind -= 1

            # store equation
            for jfront in range(frontwidth):
                kfront = flatind_sym(ifront, jfront)
                lhs[gind, jfront] = A_front[kfront]
                A_front[kfront] = 0.0
            rhs[gind] = B_front[ifront]
            B_front[ifront] = 0.0

            # eliminate from front
            pivot = lhs[gind, ifront]
            lhs[gind, ifront] = 0.0
            if presc_bool[gind] == True:
                for jfront in range(frontwidth):
                    B_front[jfront] -= lhs[gind, jfront] * presc_val[gind]
            else:
                if abs(pivot) < 1e-12:
                    raise Exception('The matrix is singular!')
                for jfront in range(frontwidth):
                    factor = lhs[gind, jfront] / pivot
                    if abs(factor) < 1e-12:
                        continue
                    for kfront in range(jfront, frontwidth):
                        lfront = flatind_sym(jfront, kfront)
                        A_front[lfront] -= factor * lhs[gind, kfront]
                    B_front[jfront] -= factor * rhs[gind]
            lhs[gind, ifront] = pivot
            front_to_glob[ifront] = 0

            # store data for backsubstitution
            glob_to_width[gind] = frontwidth
            glob_to_front[gind] = ifront
            eqpath[iEQ] = gind
            cElim += 1
            iEQ += 1

        # decrease frontwidth if possible
        for ifront in range(frontwidth - 1, -1, -1):
            if front_to_glob[ifront] != 0:
                break
            frontwidth -= 1
    topology -= 1
    return lhs, rhs, eqpath, glob_to_front, glob_to_width


if __name__ == '__main__':
    pass
    """
    from polydata.mechanics.fem import Uniform3D
    from polydata.mechanics.material.hooke import Hooke
    from dewloosh.math.linalg.frontal.path import optimal_elimination_path
    from dewloosh.math.linalg import solve
    from dewloosh.math.linalg.frontal.postproc import backsub_fr
    from polydata.core import NestedDefaultDict
    import numpy as np

    solutions = NestedDefaultDict()
    solvers = ['frontal', 'numpy', 'Gauss-Jordan']
    for solver in solvers:

        # geometry
        Lx, Ly, Lz = shape = (100, 100, 100)

        # mesh control
        nx, ny, nz = size = (5, 5, 5)

        # assembly
        assembly = Uniform3D(shape, size)
        cells = assembly.cells()
        points = assembly.points()
        coord = np.array(points)

        # material
        model = Hooke(E=2000, NU=0.3)
        for c in cells:
            c.model = model

        # essential boundary conditions
        condZ = (abs(coord[:, 2] - Lz/2) < 0.001)
        condX = (coord[:, 0] < 0.001) | (coord[:, 0] > Lx-0.001)
        condY = (coord[:, 1] < 0.001) | (coord[:, 1] > Ly-0.001)
        cond = (condX | condY) & condZ
        nIDs = np.where(cond)[0]
        if solver == 'numpy':
            p = np.array([1e12, 1e12, 1e12], dtype=np.float64)
            penalties = {nID: p for nID in nIDs}
        else:
            disps = {}
            for nID in nIDs:
                for idof in range(3):
                    disps[nID * 3 + idof] = 0.0

        # natural boundary conditions
        case = 0
        if case == 0:
            nIDs = np.where(condZ)[0]
            fz = (-10/10000)*Lx*Ly/len(nIDs)
            fp = np.array([0, 0, fz], dtype=np.float64)
            loads = {nID: fp for nID in nIDs}
        elif case == 1:
            condZ = (abs(coord[:, 2] - Lz/2) < 0.001)
            condX = (abs(coord[:, 0] - Lx/2) < 0.001)
            condY = (abs(coord[:, 1] - Ly/2) < 0.001)
            cond = condX & condY & condZ
            nIDs = np.where(cond)[0]
            loads = {nIDs[0]: np.array([0., 0., -10.0])}

        # preprocessor
        if solver == 'frontal':
            assembly.preprocess(solver=solver)
        elif solver == 'numpy':
            assembly.preprocess(solver=solver)
        elif solver == 'Gauss-Jordan':
            assembly.preprocess(solver='numpy')
        elif solver == 'Jordan':
            assembly.preprocess(solver='numpy')

        # solve
        if solver == 'frontal':
            nTOTV = len(points) * 3

            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            K, Ktopo = assembly.stiffness_matrix_frontal()

            # connectivity and topology
            adj_csr = assembly.adjacency_matrix(sparse=True, mode='node')
            topo_csr = assembly.connectivity_matrix(sparse=True,
                                                    mode='node')
            epath = optimal_elimination_path(adj_csr, topo_csr)
            #epath = np.arange(len(cells))

            # solution
            A = K
            B = RHS.reshape((nTOTV, 1))
            topology = Ktopo
            lhs, rhs, eqpath, glob_to_front, glob_to_width = \
                frontal_sym_bulk_uniform(A, topology, B,
                                         presc_bool, presc_val, epath)
            U, R = backsub_fr(lhs, rhs, presc_bool, presc_val, eqpath,
                              glob_to_front, glob_to_width)
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)
        elif solver == 'numpy':
            assembly.process(nodal_penalties=penalties, nodal_loads=loads,
                             solver=solver)
        elif solver == 'Gauss-Jordan':
            nTOTV = len(points) * 3

            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            A = assembly._K.todense()
            B = RHS.reshape((nTOTV, 1))
            U, R = solve(A, B, presc_bool, presc_val, method='Gauss-Jordan')
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)
        elif solver == 'Jordan':
            nTOTV = len(points) * 3

            # essential boundary conditions
            presc_bool = np.zeros(nTOTV, dtype=np.int64)
            presc_val = np.zeros(nTOTV, dtype=np.float64)
            for dID, dval in disps.items():
                presc_val[dID] = dval
                presc_bool[dID] = 1

            # natural boundary conditions
            RHS = np.zeros((len(points), 3), dtype=np.float64)
            for nID, f in loads.items():
                RHS[nID] = f

            # stiffness
            A = assembly._K.todense()
            B = RHS.reshape((nTOTV, 1))
            U, R = solve(A, B, presc_bool, presc_val, method='Jordan')
            assembly.point_data['dof'] = U.reshape(RHS.shape)
            assembly.point_data['res'] = R.reshape(RHS.shape)

        solutions[solver]['min'] = np.min(assembly.point_data['dof'])
        solutions[solver]['max'] = np.max(assembly.point_data['dof'])
        solutions[solver]['u[0]'] = assembly.point_data['dof'][0]

    print(solutions['frontal']['max'])
    print(solutions['numpy']['max'])
    print(solutions['Gauss-Jordan']['max'])
    """